"""SXPS data access functions.

This module provides SXPS-specific functions for downloading data,
products or more information. It allows direct access, rather than
having to carry out a query first.

Provided functions:

* getSourceDetails()
* getLightCurves()
* saveLightCurves()
* getSpectra()
* saveSpectra()
* saveSourceImages()
* getUpperLimits()
* getFailedUpperLimit()
* getDatasetDetails()
* saveDatasetImages()

"""

__docformat__ = "restructedtext en"


import os
import pandas as pd
import tempfile
import subprocess
from .. import main as base
from . import download as dl


if base.HAS_ASTROPY:
    import astropy.coordinates  # noqa

LC_BINNING = {
    "observation": "Observation",
    "obs": "Observation",
    "obsid": "Observation",
    "snapshot": "Snapshot",
    "ss": "Snapshot",
    "counts": "Counts",
}

LC_TIMEFORMAT = {
    "met": "Swift MET",
    "swift met": "Swift MET",
    "mjd": "MJD",
    "tdb": "TDB",
}

UL_BOOLCOLS = (
    "IsDetected",
    "IsRetrospectiveDetection",
    "DataFromCatalogue",
    "PileupWarning",
)

UL_FLOATCOLS = ("UpperLimit", "Rate", "RatePos", "RateNeg", "Counts", "BGCounts", "CorrectionFactor")

BAND_NAMES = base.SXPS_BAND_NAMES

tableLookup = {
    "extcatmatches": "xcorr",
}

# ----------------------------------------------------------------------
# Source functions


def _handleSourceArgs(sourceID, sourceName):
    """Internal function to handle sourceID/name arguments.

    This function receives the sourceID and sourceName passed in
    a call, ensures that only one is set, and then converts the value
    into a list if necessary.

    Parameters
    ----------

    sourceID : int or None
        The sourceID specified by the user
    sourceName : str or None
        The source name specified by the user

    Returns
    -------

    varName : str
        Either 'sourceID' or 'sourceName', the name of the argument to
            send to the API call.
    varVal : list
        A list of the sourceIDs/names supplied
    single: bool
        Whether a single value, instead of a list, was supplied.

    """
    # Exactly one of sourceID and sourceName must be set, i.e. not None.
    # I don't see an exclusive-or operator in Python, but we can fake one
    # as below. a xor b; where a and b are bools, can be done as a!=b,
    # but we want to raise when this is not the case, i.e. not xor,
    # so if a==b then we have an error. a is "sourceID is None", and b is
    # "sourceName is None".
    # So the following line basically raises an error if both or neither
    # are set.
    if (sourceID is None) == (sourceName is None):
        raise ValueError("Exactly one of `sourceID` or `sourceName` must be set.")

    varName = None
    varVal = None
    single = False

    if sourceName is not None:
        varName = "sourceName"
        if isinstance(sourceName, (list, tuple)):
            varVal = sourceName
        else:
            varVal = [sourceName]
            single = True

    else:
        varName = "sourceID"
        if isinstance(sourceID, (list, tuple)):
            varVal = sourceID
        else:
            varVal = [sourceID]
            single = True

    return (varName, varVal, single)


def _transientToSource(cat, varName, varVal, skipErrors=False, silent=False, verbose=True):
    """Convert transients to sources.

    This internal function receives the parameters created by
    _handleSourceArgs and converts the identifiers/names from transient
    handles into source handles.

    Parameters
    ----------

    cat : str
        The catalogue we are working with.

    varName : str
        Either 'sourceID' or 'sourceName', the name of the argument to
            send to the API call.

    varVal : list
        A list of the sourceIDs/names supplied

    skipErrors : bool, optional
        Whether to continue if a transAsSource results in an error as
        above (default: ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    Returns
    -------

    list
        The corrected ``varVal`` list.

    """
    if verbose:
        silent = False
    tmpSend = {"whichCat": cat, varName: varVal}
    tmp = base.submitAPICall("getSXPSTransAsSources", tmpSend, verbose=verbose, minKeys=("lookup",))
    tmp2 = varVal
    varVal = []
    for k in tmp2:
        k = str(k)
        if (k not in tmp["lookup"]) or (tmp["lookup"][k] == "MISSING"):
            if not silent:
                print(f"Transient `{k}` is not yet in LSXPS.")
            if not skipErrors:
                raise RuntimeError(f"Transient `{k}` is not yet in LSXPS.")
        varVal.append(tmp["lookup"][k])
    return varVal


def getSourceDetails(sourceID=None, sourceName=None, cat="LSXPS", silent=True, verbose=False, forceSingle=False):
    """Get the full set of information for an SXPS source.

    This obtains all of the information for a given SXPS source as
    displayed on the webpages. This duplicates some of the information
    in the Public_Sources table, but includes other information such
    as details of detections/non-detections, external catalogue macthes
    and so forth.

    sourceID or sourceName can be set - an error will be raised if both
    (or neither) is set - and can either be a single value, or a list.
    If a single value, then this function returns a single dict
    containing the details of the requested source. If a list, then
    the return value will be a list of dicts.

    Because LSXPS is a dynamic catalogue, source definitions can change
    and a given source can be superseded; indeed, a source can be split
    and so superseded by multiple sources (see the LSXPS documentation
    for more details). If a requested source has been superseded by a
    single newer source then the details of that newer source will be
    returned; you will be able to tell that this has happened because
    the source ID or name will not match your request. In your requested
    source has been split into multiple sources, the return dict will
    have the single key 'newerSources' which is a list of the
    superseding sources; each list entry is a dict giving basic
    information about the superseding source.

    Parameters
    ----------

    sourceID : int or list or tuple, optional
        The SXPS source ID.

    sourceName : str or list or tuple, optional
        The SXPS source IAUName.

    cat : str
        Which SXPS catalogue the source is in.

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    forceSingle : bool
        For internal usage. Top secret.

    Returns
    -------

    list or dict
        The source info dict, or a list of dicts if a list of sources
        was supplied.

    """
    if verbose:
        silent = False  # noqa
    varName, varVal, single = _handleSourceArgs(sourceID, sourceName)

    if forceSingle:
        single = True
        if len(varVal) != 1:
            raise RuntimeError("Internal error via transAsSource. Please report this as a bug.")

    pdl1 = ("Detections", "NonDetections")
    pdl2 = ("Observations", "Stacks")
    # The argument has now been made into a list, so we can go through
    # that list and submit the request to the API.
    # We will store the results in ret
    ret = {}

    sendData = {"whichCat": cat}
    for i in varVal:
        if verbose:
            print(f"Getting data for {varName} = `{i}`")
        sendData[varName] = i
        # We can just accept this result, as the only key we have to have is "OK", and
        # if that wasn't set we will already have hit an error.
        tmp = base.submitAPICall("getSXPSSourceInfo", sendData, verbose=verbose)

        if "data" in tmp:
            ret[i] = tmp["data"]
        else:
            ret[i] = tmp

        for x in pdl1:
            if x in ret[i]:
                for y in pdl2:
                    if y in ret[i][x]:
                        ret[i][x][y] = pd.DataFrame(ret[i][x][y])
        if "CrossMatch" in ret[i]:
            ret[i]["CrossMatch"] = pd.DataFrame(ret[i]["CrossMatch"])

    if single:
        return ret[varVal[0]]
    else:
        return ret


def getLightCurves(
    sourceID=None,
    sourceName=None,
    cat="LSXPS",
    bands="all",
    binning=None,
    timeFormat=None,
    returnData=False,
    saveData=True,
    silent=True,
    verbose=False,
    transient=False,
    transAsSource=False,
    skipErrors=False,
    **kwargs,
):
    """Download a light curve, or set of light curves from SXPS.

    This will retrieve a light curve, or set of light curves, for
    object(s) in an SXPS catalogue, converting them into a
    set of pandas DataFrames.

    Some notes about what is returned, when I've worked it out.

    sourceID or sourceName can be set - an error will be raised if both
    (or neither) is set - and can either be a single value, or a list.
    If a single value, then this function returns a single dict
    containing the details of the requested source. If a list, then the
    return value will be a dict of dicts, indexed by the sourceID or
    name, whichever you supplied.

    This function can also be used to get light curves for transients,
    by setting the ``transient`` parameter to ``True``. In this case the
    ID or name supplied should be the transient identifier. Transient
    light curves are only available in the total band, with the time
    axis in MJD, and with the definitions of bins and upper limits fixed
    but with an extra binning option ('counts'). Supplied arguments that
    are not supported for transients will be ignored, and warnings will
    only appear if ``silent`` is ``False``.

    Transient sources will appear in LSXPS, with some delay, and so you
    can access the normal LSXPS light curve for a source via its
    transient identifier, by specifiying ``transAsSource=True``, however
    if the transient is not yet in LSXPS, this will return an error.

    Because LSXPS is a dynamic catalogue, source definitions can change
    and a given source can be superseded; indeed, a source can be split
    and so superseded by multiple sources (see the LSXPS documentation
    for more details). If a requested source has been superseded by a
    single newer source then the details of that newer source will be
    returned; you will be able to tell that this has happened because
    the source ID or name will not match your request. If your requested
    source has been split into multiple sources, the return dict will
    have the single key 'newerSources' which is a list of the
    superseding sources; each list entry is a dict giving basic
    information about the superseding source.

    Parameters
    ----------

    sourceID : int or list or tuple, optional
        The SXPS source ID.

    sourceName : str or list or tuple, optional
        The SXPS source IAUName.

    cat : str, optional
        Which SXPS catalogue the source is in (default: 'LSXPS').

    bands : str or list, optional
        Either 'all', or a list of labels of the energy bands to
        retrieve ('total', 'soft', 'medium', 'hard', 'HR1', 'HR2');
        (default: 'all')

    binning : str
        Which of the two binning methods you want: either
        'observation' or 'snapshot'

    timeFormat : str
        Which units to use for the time axis. Must be 'met' or 'tdb'
        or 'mjd'.

    returnData : bool, optional
        Whether to return the organised light curve data from this
        function (default: ``False``).

    saveData : bool, optional
        Whether to save the light curve data to disk (default:
        ``True``).

    transient : bool, optional
        Whether the object requested is from the LSXPS transient list,
        rather than source list (i.e. the name/ID is a transient name/
        ID) (default: ``False``).

    transAsSource : bool, optional
        Only relevant if ``transient=True``, this means that the object
        identifier supplied is a transient, but the desired light curve
        should be taken from the LSXPS data corresponding to this
        transient. If the transient is not (yet) in LSXPS then this will
        result in an error (default: ``False``).

    skipErrors : bool, optional
        Whether to continue if a transAsSource results in an error as
        above (default: ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    **kwargs : dict, optional
        Any options to pass to ``saveLightCurve()`` and also extra
        options controlling your light curves, to pass to the API. These
        are:

        * showObs : bool, optional
            Whether the obsID should be included as a column.

        * hideBadPup : bool, optional
            Whether bins with the "pupWarn" flag set should be
            rejected.

        * mergeAliases : bool
            Whether to include points from sources identified as
            aliases.

        * getAllTypes : bool
            Whether to get all possible data types, and store
            as separate entities. This overrides all of the following
            arguments.

        * groupRates : bool
            Whether all detections (blind and retrospective)
            should be grouped into the same light curve entity.

        * groupULs : bool
            Whether all upper limits (non-detection, and
            retrospective only) should be grouped into the same light
            curve entity.

        * groupHRss : bool
            Whether all HRs (blind and retrospective dets)
            should be grouped into the same light curve entity.

        * retroAsUL  : bool
            Whether to return upper limits, instead of
            count-rate bins, for retrospective detections

        * nondetAsUL : bool
            Whether to return upper limits, instead of
            count-rate bins, for non detections

        * getRetroHR : bool
            Whether to include HRs from retrospective detections.


    Returns
    ------

    dict :

        If a list of sources was supplied, then the dictionary will have
        keys corresponding to the sourceID or sourceName, depending on
        which this function was called with. Each of the entries in this
        dict is itself a dict corresponding to the results for a single
        source.

        If a sourceID/name was supplied not as a list, then the dict
        returned is that containing the results for that source.

        The contents of this dict range significantly, depending on the
        details of your request. The following keys will always exist:

        * TimeFormat: Confirming which time system the time axis uses.

        * Binning: Whether observation or snapshot binning was selected.

        * Datasets: A list of the keys corresponding to the datasets
          (that is, individual light curves) returned.

        Then there is an entry per light curve *returned*. NB this may
        not match the light curves requested, as empty light curves are
        not returned.

        The names of these datasets depends upon the arguments supplied
        above, and differs for light curves and hardness ratios.

        For light curves, the general form is:

        ${band}_${what}

        ${band} is one of "total", "soft", "medium", "hard".

        ${what} is more complex, as it depends upon the "group" options
        supplied in the arguments. BY default, groupRates and groupULs
        are True, in which case ${what} will be:

        * rates -- for the count rates with 1-sigma errors *  UL    --
          for the 3-sigma upper limits.

        However if groupRates is False then instead there will be up to
        three "rates" entries for ${what}:

        * "blind_rates" - for cases where a blind detection of the
          source exists.

        * "retro_rates" - for cases where retrospective detections of
          the source exists.

        * "nondet_rates" - for cases where the source was not detected,
          blindly or retrospecitvely.

        Similarly, if groupULs is False then there will up to two "UL"
        entries for ${what}:

        * "retro_UL" - for cases where only a retrospective detection
          exists.

        * "nondet_UL" - for cases where the source was not detected,
          blindly or retrospecitvely.

        Normally, not all of these will exist, as you can specify
        whether retrospective and non-detections are reported as upper
        limits or as count-rate bins, using the retroAsUL and nondetAsUL
        arguments; by default the former is False and the latter True;
        that is, all detections (blind or retrospective) are reported as
        count-rate bins and non-detections as upper limits.

        If the parameters "getAllTypes" is set to True then both upper
        limits AND count-rate bins will be produced for the
        retrospective and non-detections, and all no grouping is
        applied.

        Note that blind detections are always reported with count-rate
        bins.

    """
    if verbose:
        silent = False

    varName, varVal, single = _handleSourceArgs(sourceID, sourceName)

    # Now handle the other arguments.
    if transient and not transAsSource:
        if not silent:
            if len(kwargs) > 0:
                print("WARNING: kwargs are ignored for transient light curves")
            if (bands is not None) and (bands.lower() != "all"):
                print("WARNING: 'bands' is ignored for transient light curves")
            if (timeFormat is not None) and (timeFormat.lower() != "mjd"):
                print("WARNING: 'timeFormat' is ignored for transient light curves")

    sendData = {
        "whichCat": cat,
        "bands": bands,
        "binning": binning,
        "timeFormat": timeFormat,
    }

    if transient and not transAsSource:
        sendData["transient"] = True

    # May need to convert transientIDs into sourceIDs
    if transient and transAsSource:
        varVal = _transientToSource(cat, varName, varVal, skipErrors=skipErrors, silent=silent, verbose=verbose)

    # And there are other things that may have been set in kwargs.
    # These are all bools. Start with the universal ones, and getAllTypes
    if transAsSource or not transient:
        tmp = ("showObs", "hideBadPup", "mergeAliases", "getAllTypes")
        for t in tmp:
            if t in kwargs.keys():
                if not isinstance(kwargs[t], bool):
                    raise RuntimeError(f"`{t}` should be boolean")
                sendData[t] = kwargs[t]

        # Now the grouping and retrieval things.
        getAll = False
        if "getAllTypes" in sendData:
            getAll = sendData["getAllTypes"]

        # Now handle all the grouping things
        tmp = (
            "groupRates",
            "groupULs",
            "groupHRss",
            "retroAsUL",
            "nondetAsUL",
            "getRetroHR",
        )
        for t in tmp:
            if t in kwargs:
                if getAll:
                    if not silent:
                        print(f"WARNING: Ignoring parameter {t} as getAllTypes = True")
                    continue
                if not isinstance(kwargs[t], bool):
                    raise RuntimeError(f"`{t}` should be boolean")
                sendData[t] = kwargs[t]

    # If we had a list, then we will return a dict, where each API result
    # is an entry, and the supplied keys (IDs or names) are the dict keys.
    # If only a single value was supplied, we will return only the first
    # entry from this.
    # NOTE: A list with a single entry is returned as a dict with a single
    # entry.

    binLookup = {"obslc": "Observation", "sslc": "Snapshot", "counts": "Counts"}
    ret = {}

    for i in varVal:
        if verbose:
            print(f"Getting data for {varName} = `{i}`")
        sendData[varName] = i

        # We can just accept this result, as the only key we have to have is "OK", and
        # if that wasn't set we will already have hit an error.
        tmp = base.submitAPICall("getSXPSLC", sendData, verbose=verbose)

        if tmp["Binning"] in binLookup:
            tmp["Binning"] = binLookup[tmp["Binning"]]

        if "Datasets" in tmp:
            ret[i] = dl._handleLightCurve(tmp, silent, verbose)
        else:
            ret[i] = tmp

    if saveData:
        d = ret
        if single:
            d = ret[varVal[0]]
        saveLightCurves(d, silent=silent, verbose=verbose, **kwargs)

    if returnData:
        if single:
            return ret[varVal[0]]
        else:
            return ret


def saveLightCurves(
    data,
    destDir="lc",
    whichSources="all",
    subDirs=True,
    **kwargs,
):
    """Save light curves to text files.

     This will save the light curve data supplied via the `data`
    parameter to disk. By default, this will save all of the data in the
    supplied object but you can request to save only specific subsets.
    Parameters
    ----------

    data : dict
        The light curve data, either a light curve `dict` or a `dict` of
        light curve `dict`s.

    destDir : str, optional
        The directory in which to save the files (default: 'lc').

    whichSources : list or str, optional
        A list of the sources to save. Should be keys of the data
        object, or 'all' (default: 'all').

    subDirs : bool, optional
        This argument is ignored if ``data`` is a light curve ``dict``,
        rather than a ``dict`` of light curve ``dict``s.

        Whether to save the light curves for the separate GRBs in
        subdirectories of ``destDir``, with directories named by the
        keys in ``data``. If ``False``, the keys will be
        prepended to the file names, since the files for different GRBs
        otherwise all have the same names (default: ``True``).

    **kwargs : dict, optional
        Other arguments to pass to _saveLightCurveFromDict()


    """
    if "silent" in kwargs:
        silent = kwargs["silent"]
    else:
        silent = True
    if "verbose" in kwargs:
        verbose = kwargs["verbose"]
    else:
        verbose = False

    if verbose:
        silent = False

    usePrefix = True
    # Need to check if we have a light curve dict, or a set of dicts.
    if "Datasets" in data:  # LC dict
        usePrefix = False  # Do not prepend this fake sourceID to the file name.
        tmp = {"mySource": data}
        data = tmp
        # If we had an lc dict, then we don't use subDirs (NB a list with one entry does use them)
        subDirs = False

    # Create the output dir, if needed.

    base._createDir(destDir, silent=silent, verbose=verbose)

    # Which sources am I saving?
    if (whichSources is None) or (whichSources == "all"):
        whichSources = data.keys()

    # For SXPS objects we add the time or binning to the extension unless the user says no
    if "timeFormatInFname" not in kwargs:
        kwargs["timeFormatInFname"] = True
    if "binningInFname" not in kwargs:
        kwargs["binningInFname"] = True

    # prefix should not be in this dict, so remove if it is.
    kwargs.pop("prefix", None)

    # Now start saving light curves, one source at a time
    for source in whichSources:
        if source not in data:
            raise ValueError(f"{source} is not in the light curve list.")

        if "newerSources" in data[source]:
            if not silent:
                print(f"Skipping source {source} as is has been superseded by multiple sources.")
            continue

        path = destDir
        prefix = ""
        if subDirs:
            path = f"{destDir}/{source}"
        elif usePrefix:
            prefix = f"{source}_"

        dl._saveLightCurveFromDict(data[source], destDir=path, prefix=prefix, **kwargs)


def getSpectra(
    sourceID=None,
    sourceName=None,
    cat="LSXPS",
    returnData=False,
    saveData=True,
    transient=False,
    transAsSource=False,
    specType=None,
    skipErrors=False,
    silent=True,
    verbose=False,
    **kwargs,
):
    """Download a spectrum, or set of spectra from SXPS.

    This will retrieve the information about a spectral fit for
    object(s) in an SXPS catalogue, returning a dictionary of the
    details, which also includes the URL of a tar file containing the
    actual spectrum data, and URLs of gif images of the fitted spectra.

    sourceID or sourceName can be set - an error will be raised if both
    (or neither) is set - and can either be a single value, or a list.
    If a single value, then this function returns a single dict
    containing the details of the requested source. If a list, then the
    return value will be a list of dicts.

    This function can also be used to get spectra for transients,
    by setting the ``transient`` parameter to ``True``. In this case the
    ID or name supplied should be the transient identifier. In this case
    you must also suppled the ``specType`` parameter to specify which
    of the two transient spectra you want to retrieve.

    Transient sources will appear in LSXPS, with some delay, and so you
    can access the normal LSXPS light curve for a source via its
    transient identifier, by specifiying ``transAsSource=True``, however
    if the transient is not yet in LSXPS, this will return an error.

    Because LSXPS is a dynamic catalogue, source definitions can change
    and a given source can be superseded; indeed, a source can be split
    and so superseded by multiple sources (see the LSXPS documentation
    for more details). If a requested source has been superseded by a
    single newer source then the details of that newer source will be
    returned; you will be able to tell that this has happened because
    the source ID or name will not match your request. If your requested
    source has been split into multiple sources, the return dict will
    have the single key 'newerSources' which is a list of the
    superseding sources; each list entry is a dict giving basic
    information about the superseding source.

    Parameters
    ----------

    sourceID : int or list or tuple
        The SXPS source ID.

    sourceName : str or list or tuple
        The SXPS source IAUName.

    cat : str
        Which SXPS catalogue the source is in.

    returnData : bool, optional
        Whether to return the organised light curve data from this
        function (default: ``False``).

    saveData : bool, optional
        Whether to save the light curve data to disk (default:
        ``True``).

    transient : str, optional
        Whether the object requested is from the LSXPS transient list,
        rather than source list (i.e. the name/ID is a transient name/
        ID) (default: ``False``).

    transAsSource : bool, optional
        Only relevant if ``transient=True``, this means that the object
        identifier supplied is a transient, but the desired light curve
        should be taken from the LSXPS data corresponding to this
        transient. If the transient is not (yet) in LSXPS then this will
        result in an error (default: ``False``).

    specType : str, optional
        Required if ``transient`` is ``True``, and ``transAsSource`` is
        ``False``, this specifies whether to get the spectrum created
        only from the discovery dataset (set to 'discovery') or that
        made from all data since discovery ('full'). If it is 'both'
        then the returned ``dict`` has keys 'Discovery' and 'Full'
        containing the respective spectra.

    skipErrors : bool, optional
        Whether to continue if a transAsSource results in an error as
        above (default: ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    **kwargs : dict, optional
        Any arguments to pass to ``saveSpectra()``

    Returns
    ------

    dict

        An API-standard spectrum dict.

    """
    if verbose:
        silent = False

    varName, varVal, single = _handleSourceArgs(sourceID, sourceName)

    # Now handle the other arguments.

    # Now handle the other arguments.

    sendData = {"whichCat": cat}
    if transient and not transAsSource:
        sendData["transient"] = True

    # May need to convert transientIDs into sourceIDs
    if transient:
        if transAsSource:
            varVal = _transientToSource(cat, varName, varVal, skipErrors=skipErrors, silent=silent, verbose=verbose)
        else:
            if specType is None:
                raise RuntimeError("`specType` must be set when getting a transient spectrum.")
            else:
                sendData["specType"] = specType

    ret = {}
    for i in varVal:
        if verbose:
            print(f"Getting data for {varName} = `{i}`")
        sendData[varName] = i

        # We can just accept this result, as the only key we have to have is "OK", and
        # if that wasn't set we will already have hit an error.
        ret[i] = base.submitAPICall("getSXPSSpectrumData", sendData, verbose=verbose)

    if saveData:
        saveAsTrans = transient and not transAsSource and specType.lower() == "both"
        d = ret
        if single:
            d = ret[varVal[0]]
        saveSpectra(d, silent=silent, verbose=verbose, transient=saveAsTrans, **kwargs)

    if returnData:
        if single:
            return ret[varVal[0]]
        else:
            return ret


def saveSpectra(data, destDir="spec", whichSources=None, transient=False, silent=True, verbose=False, **kwargs):
    """Save the spectral data to disk.

    This will download and save the spectral files from a previous
    ``getSpectra()`` call.

    Parameters
    ----------

    data : dict
        The object returned by ``getSXPSSpectra()``

    destDir : str, optional
        The directory in which to save the spectra (default: "spec").

    whichSources : list
        A list of the sources to save. Should be keys of the data
        object.

    transient : bool, optional
        Whether the spectra are transient spectra (default: ``True``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    **kwargs : dict, optional
        Arguments to pass to ``download.dl._saveSpectrum()``

    """
    if verbose:
        silent = False

    # Little hack needed here to handle the case that we have only a single spectrum, not a dict of LCs
    # We will know this because it will have the 'Datasets' key
    single = False
    if "rnames" in data:
        tmp = {"mySource": data}
        data = tmp
        single = True

    # Create the output dir, if needed.
    base._createDir(destDir, silent=silent, verbose=verbose)

    # Which sources am I saving?
    if (whichSources is None) or (whichSources == "all"):
        whichSources = data.keys()

    # Now start saving spectra, one source at a time
    for source in whichSources:
        if source not in data:
            raise ValueError(f"{source} is not in the spectra available.")
        if "NoSpectrum" in data[source]:
            if not silent:
                print(f"Source `{source}` has no spectra.")
            continue

        if "newerSources" in data[source]:
            if not silent:
                print(f"Skipping source {source} as is has been superseded by multiple sources.")
            continue

        path = destDir
        if not single:
            path = f"{destDir}/{source}"
            base._createDir(path, silent=silent, verbose=verbose)

        print(f"Transient {transient}")
        if transient:
            for i in ("Full", "Discovery"):
                if i in data[source]:
                    if "NoSpectrum" in data[source][i]:
                        if not silent:
                            print(f"Source `{source}` has no `{i}` spectrum.")
                        continue
                    else:
                        fpath = f"{path}/{i}"
                        base._createDir(fpath, silent=silent, verbose=verbose)
                        dl._saveSpectrum(data[source][i], destDir=fpath, silent=silent, verbose=verbose, **kwargs)
        else:
            dl._saveSpectrum(data[source], destDir=path, silent=silent, verbose=verbose, **kwargs)


def saveSourceImages(
    sourceID=None,
    sourceName=None,
    cat="LSXPS",
    bands="all",
    destDir="images",
    subDirs=True,
    clobber=False,
    skipErrors=False,
    transient=False,
    transAsSource=False,
    silent=True,
    verbose=False,
):
    """Download the thumbnail images of the sources.

    This will retrieve and save thumbnail images of the specified
    object(s) in an SXPS catalogue.

    sourceID or sourceName can be set - an error will be raised if both
    (or neither) is set - and can either be a single value, or a list.

    Because LSXPS is a dynamic catalogue, source definitions can change
    and a given source can be superseded; indeed, a source can be split
    and so superseded by multiple sources (see the LSXPS documentation
    for more details). If a requested source has been superseded by a
    single newer source then the details of that newer source will be
    returned; you will be able to tell that this has happened because
    the source ID or name will not match your request. If your requested
    source has been split into multiple sources, the return dict will
    have the single key 'newerSources' which is a list of the
    superseding sources; each list entry is a dict giving basic
    information about the superseding source.

    Parameters
    ----------

    BLAH

    cat : str
        Which SXPS cat the source is in.

    bands : str or list, optional
        Which bands' images to retrieve. As well as the normal SXPS
        bands, this can include 'Expmap' to get the exposure map image
        (default: 'all').

    subDirs : bool, optional
        Whether to save each dataset's data into its own subdirectory
        (default: ``True``).

    clobber : bool
        Whether to overwrite files if they exist.

    skipErrors : bool, optional
        If an error occurs saving a file, do not raise a RuntimeError
        but simply continue to the next file (default ``False``).

    transient : str, optional
        Whether the object requested is from the LSXPS transient list,
        rather than source list (i.e. the name/ID is a transient name/
        ID) (default: ``False``).

    transAsSource : bool, optional
        Only relevant if ``transient=True``, this means that the object
        identifier supplied is a transient, but the desired light curve
        should be taken from the LSXPS data corresponding to this
        transient. If the transient is not (yet) in LSXPS then this will
        result in an error (default: ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    """
    if verbose:
        silent = False

    varName, varVal, single = _handleSourceArgs(sourceID, sourceName)

    sendData = {"whichCat": cat, "bands": bands}

    if transient and not transAsSource:
        sendData["transient"] = True

    # May need to convert transientIDs into sourceIDs
    if transient and transAsSource:
        varVal = _transientToSource(cat, varName, varVal, skipErrors=skipErrors, silent=silent, verbose=verbose)

    # Create the output dir, if needed.
    base._createDir(destDir, silent=silent, verbose=verbose)
    for i in varVal:
        if verbose:
            print(f"Getting images for {varName} = `{i}`")
        sendData[varName] = i

        # We can just accept this result, as the only key we have to have is "OK", and
        # if that wasn't set we will already have hit an error.
        tmp = base.submitAPICall("getSXPSSourceImages", sendData, verbose=verbose)

        path = destDir
        prefix = None
        if subDirs:
            path = f"{path}/{i}"
            base._createDir(path, silent=silent, verbose=verbose)
        else:
            prefix = f"{i}_"
        for b in BAND_NAMES:
            if b in tmp:
                url = tmp[b]
                if (url == "NOTFOUND") and (not silent):
                    print(f"No image exists for the {b} band for `{i}`.")
                else:
                    ok = dl._saveURLToFile(
                        url,
                        path,
                        name=f"{b}.png",
                        prefix=prefix,
                        clobber=clobber,
                        silent=silent,
                        verbose=verbose,
                    )
                    if not ok:
                        if skipErrors:
                            continue
                        raise RuntimeError(f"Failed to save {url} to {path}/")
        if "Expmap" in tmp:
            url = tmp["Expmap"]
            if (url == "NOTFOUND") and (not silent):
                print(f"No exposure map image exists for `{i}`.")
            else:
                ok = dl._saveURLToFile(
                    url,
                    path,
                    name="Expmap.png",
                    prefix=prefix,
                    clobber=clobber,
                    silent=silent,
                    verbose=verbose,
                )
                if not ok:
                    if skipErrors:
                        continue
                    raise RuntimeError(f"Failed to save {url} to {path}/")


# ----------------------------------------------------------------------
# Transient functions


def getTransientDetails(
    sourceID=None, sourceName=None, cat="LSXPS", transAsSource=False, skipErrors=False, silent=True, verbose=False
):
    """Get the full set of information for an SXPS transient.

    This obtains all of the information for a given SXPS transient as
    displayed on the webpages. This duplicates some of the information
    in the Public_Transients table, but includes other information.

    sourceID or sourceName can be set - an error will be raised if both
    (or neither) is set - and can either be a single value, or a list.
    If a single value, then this function returns a single dict
    containing the details of the requested source. If a list, then
    the return value will be a list of dicts.

    Parameters
    ----------

    sourceID : int or list or tuple, optional
        The SXPS source ID.

    sourceName : str or list or tuple, optional
        The SXPS source IAUName.

    cat : str
        Which SXPS catalogue the source is in.

    transAsSource : bool, optional
        If set, then this will convert the transientID to a sourceID
        (if possible) and then call getSourceDetails instead.

    skipErrors : bool, optional
        Whether to continue if converting to sourceID results in an
        error (default: ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    Returns
    -------

    list or dict
        The source info dict, or a list of dicts if a list of sources
        was supplied.

    """
    if verbose:
        silent = False

    varName, varVal, single = _handleSourceArgs(sourceID, sourceName)
    if transAsSource:
        varVal = _transientToSource(cat, varName, varVal, skipErrors=skipErrors, silent=silent, verbose=verbose)
        args = {varName: varVal, "cat": cat, "silent": silent, "verbose": verbose, "forceSingle": single}
        return getSourceDetails(**args)

    ret = {}

    sendData = {"whichCat": cat}
    for i in varVal:
        if verbose:
            print(f"Getting data for {varName} = `{i}`")
        sendData[varName] = i
        # We can just accept this result, as the only key we have to have is "OK", and
        # if that wasn't set we will already have hit an error.
        tmp = base.submitAPICall("getSXPSTransInfo", sendData, verbose=verbose)
        ret[i] = tmp

    if single:
        return ret[varVal[0]]
    else:
        return ret


# ----------------------------------------------------------------------
# Upper limit functions


def getUpperLimits(
    position=None,
    name=None,
    RA=None,
    Dec=None,
    cat="LSXPS",
    bands=("total",),
    whichData="deepest",
    timeFormat="cal",
    sigma=3.0,
    detectionsAsRates=False,
    detThresh=None,
    skipDetections=False,
    useCatValues=False,
    pandas=True,
    silent=True,
    verbose=False,
):
    """Find upper limit(s) for a specified object/position.

    This function returns flux information for a specific location on
    the sky. Its default use is to provide a single, 0.3-10 keV upper
    limit corresponding to the deepest observation available, however
    there is significant flexibility in what you can request, as
    demonstrated by the arguments below.

    You must supply 'position', 'name' or both 'RA' & 'Dec', but only
    one of these (e.g. you cannot give a 'position' and RA/Dec).

    Parameters
    ----------

    position : str, optional
        A string to be parsed as a free-form RA/Dec position.

    name : str, optional
        A name to be resolved into coordinates

    RA : float or astropy.coordinates.Angle, optional
        The RA at which to calculate the upper limit(s).

    Dec : float or astropy.coordinates.Angle, optional
        The declination at which to calculate the upper limit(s).

    cat : str
        Which SXPS catalogue the source is in.

    bands : str or list, optional
        Either 'all', or a list of labels of the energy bands to
        retrieve ('total', 'soft', 'medium', 'hard')
        (default: ['total']).

    whichData : str, optional
        Which datasets to use for the flux measurement, either:
        'deepest' or 'all' (default: 'deepest').

    timeFormat : str, optional
       What format/system to use for the timestamps, either: 'cal',
       'met', 'mjd', 'tdb' (default: 'cal').

    sigma : float, optional
        The confidence level for upper limits, in Gaussian sigmas
        (default: 3).

    detectionsAsRates : bool, optional
        Supply count-rate (and 1-sigma errors) in cases that the source
        is deemed 'detected' (that is, the n-sigma lower-limit is >0;
        where n is set by the detThresh parameter) (default: ``False``).

    detThresh : float, optional
        The significance level to be used for deciding if a flux
        measurement corresponds to a detection or not. If ``None``, this
        will default to sigma (default: ``None``).

    skipDetections : bool, optional
        Skip datasets in which a source is detected at / close to the
        position specified (default: ``False``).

    useCatValues : bool, optional
        If, in a given dataset, the source is in the catalogue in any
        band, then return the values from the catalogue instead of
        recalculating them (default, ``False``).

    pandas : bool, optional
        Whether to convert the result to a pandas DataFrame
        (default: ``True``).

    silent : bool, optional
        Whether to suppress all output (default ``True``).

    verbose : bool, optional
        Whether to write verbose output (default ``False``).

    Return
    ------

    dict
        This returns a dict, with the following keys (note, not all
        keys will always be present).

        * ``ULData`` - the upper limit data
        * ``DetData`` - Details of SXPS sources near the input position.
        * ``Columns`` - The list of columns returned in ULData (not
            present if ``pandas`` is ``True``).

        If you supplied the ``position`` or ``name`` argument then the
        following (standard) keys will also be present:

        * ``ResolvedRA`` The RA (decimal degrees) actually used.
        * ``ResolvedDec`` The dec (decimal degrees) actually used.
        * ``ResolvedInfo`` Text information about the coordinates.

        If the job timed out, the key ``TIMEOUT`` will be present, with
        a unique identifier that can be used to call
        ``getFailedUpperLimit()``.

        ``DetData`` will only be present if the position you requested
        is close to a source in the catalogue searched, and thus may
        serve as a warning. This will be a list, one entry per matching
        source, with keys ``Distance`` and ``SourceName``. The latter is
        self-explanatory, the former is the distance (arcsec) between
        your input position, and the source.

        ``ULData`` is the main data returned, giving the flux
        measurements your query returned. The exact contents of this
        depend both on what arguments you provided, and the values
        found when calculating the upper limits. By default, this object
        is a pandas ``DataFrame``, with one row per dataset for which
        results were obtained. If you set ``pandas=False`` then it will
        be a list (one entry per dataset) with each entry being a
        ``dict``.

        The possible keys/columns that will be in this ``ULData`` object
        are below. **Note** For brevity, only the total band entries
        are given, but for everything starting ``Total_`` there may also
        be ``Soft_``, ``Medium_`` and ``Hard_`` if you requested those
        bands.

        * ``StartTime`` - The start time of the dataset, in your
            requested format.
        * ``StopTime`` - The end time of the dataset, in your
            requested format.
        * ``ObsID`` - The observation identifier.
        * ``SourceExposure`` - The exposure (in s) at the source
            position.
        * ``ImageExposure`` - The on-axis exposure (in s) of the
            dataset.
        * ``Total_UpperLimit`` - The upper limit in the total band.
        * ``Total_Rate`` - The count rate in the total band.
        * ``Total_RatePos`` - The 1-sigma positive error on the count
            rate in the total band.
        * ``Total_RateNeg``- The 1-sigma negative error on the count
            rate in the total band.
        * ``Total_Counts`` - The measured counts in the extraction
            region in the total band.
        * ``Total_BGCounts`` - The predicted background counts in the
            extraction region in the total band.
        * ``Total_CorrectionFactor`` - The count-rate correction factor
            (correcting for vignetting, bad columns, etc) in the total
            band.
        * ``Total_IsDetected`` - Whether there was a blind detection of
            a source at your requested position in this dataset in the
            total band.
        * ``Total_IsRetrospectiveDetection`` - Whether your input
            position corresponds to an SXPS source for which there is
            a retrospective detection in this dataset in the total band.
        * ``Total_DataFromCatalogue`` - Whether these total-band data
            were taken from the catalogue, rather than being calculated
            afresh.
        * ``Total_PileupWarning`` - Whether the total-band data have
            a warning flag indicating that the pile up correction (and
            hence count-rate values) may be wrong.

    """
    if verbose:
        silent = False

    # Handle the arguments. Do not need to sanity check all arguments,
    # as this can be done by the back end and it makes sense to do it
    # just once, but we do need to handle a few things. First, put most
    # of the options into the dict to send:
    sendData = {
        "catalogue": cat,
        "bands": bands,
        "whichData": whichData,
        "timeFormat": timeFormat,
        "sigma": sigma,
        "detectionsAsRates": detectionsAsRates,
        "skipDetections": skipDetections,
        "useCatValues": useCatValues,
    }

    # 1) Do we have name, coordinates, or position?
    gotPosData = False
    if position is not None:
        sendData["searchName"] = position
        gotPosData = True
    if name is not None:
        if gotPosData:
            raise RuntimeError("You can supply position OR name, not both.")
        sendData["searchName"] = name
        gotPosData = True
    if (RA is not None) and (Dec is not None):
        if gotPosData:
            raise RuntimeError("You can supply position/name, or RA & Dec not both.")
        if isinstance(RA, astropy.coordinates.Angle):
            RA = float(RA.deg)
        elif isinstance(RA, str):
            tmp = astropy.coordinates.Angle(RA)
            RA = float(tmp.deg)
        if isinstance(Dec, astropy.coordinates.Angle):
            Dec = float(Dec.deg)
        elif isinstance(Dec, str):
            tmp = astropy.coordinates.Angle(Dec)
            Dec = float(tmp.deg)
        sendData["RA"] = RA
        sendData["Dec"] = Dec

    # 2) detThresh:
    if detThresh is None:
        detThresh = sigma

    if not silent:
        print("Submitting upper limit request; this may take a few moments.")
    tmp = base.submitAPICall("getSXPSUL", sendData, verbose=verbose)

    if verbose:
        print("Upper limit retrieved.")

    if pandas:
        _handleUL(tmp)
        tmp.pop("Columns", None)
    tmp.pop("OK", None)

    if ("TIMEOUT" in tmp) and (not silent):
        print("The upper limit calculation timed out.")
    return tmp


def getFailedUpperLimit(id, pandas=True, silent=True, verbose=False):
    """Try to get an upper limit that timed out.

    Upper limit requests called by ``getUpperLimits()`` can time out,
    in which case the dict returned by that function will contain the
    key ``TIMEOUT`` with a unique identifier as the value. It may be
    that the upper limit job still completed, but after the back-end
    timeout was hit. This function allows to to retrieve an upper limit
    in that event.

    **NOTES**

    1. Upper limit jobs are only allocated a short maximum runtime on
        the server, thus some jobs may timeout and never complete. This
        timeout is a server setting and subject to change, so cannot be
        documented here, but as a general rule, if it has been more than
        20 minutes and you still get timeout warnings, your job has
        probably been killed.

    2. When upper limit data are retrieved, by this function or by
        ``getUpperLimits()``, the files are deleted off the server, i.e.
        they can only be downloaded successfully once.

    If ``getUpperLimits()`` returned a timeout, then you can wait a
    short time and then call this function, passing as its argument the
    id string returned in the ``TIMEOUT`` key of the original request.
    This will either return another TIMEOUT warning, or the requested
    upper limit.

    Parameters
    ----------

    id : str
        The identification string for your timed out upper limit.

    pandas : bool, optional
        Whether to convert the result to a pandas DataFrame
        (default: ``True``).

    silent : bool, optional
        Whether to suppress all output (default ``True``).

    verbose : bool, optional
        Whether to write verbose output (default ``False``)..

    """
    if verbose:
        silent = False

    if not isinstance(id, str):
        raise ValueError("`id` must be a string.")

    sendData = {
        "ULID": id,
    }

    tmp = base.submitAPICall("getSXPSUL", sendData, verbose=verbose)
    if pandas:
        _handleUL(tmp)

    tmp.pop("OK", None)

    if ("TIMEOUT" in tmp) and (not silent):
        print("The upper limit calculation timed out.")
    return tmp


def _handleUL(data):
    """Internal function to convert UL data to pandas"""

    if "ULData" in data:
        data["ULData"] = pd.DataFrame(data["ULData"], columns=data["Columns"])
        for c in ("SourceExposure", "ImageExposure"):
            if c in data["Columns"]:
                data["ULData"][c] = data["ULData"][c].apply(lambda x: float(x))
        for b in BAND_NAMES:
            for col in UL_BOOLCOLS:
                c = f"{b}_{col}"
                if c in data["Columns"]:
                    data["ULData"][c] = data["ULData"][c].apply(lambda x: bool(x))
            for col in UL_FLOATCOLS:
                c = f"{b}_{col}"
                if c in data["Columns"]:
                    data["ULData"][c] = data["ULData"][c].apply(lambda x: float(x))

    if "DetData" in data:
        data["DetData"] = pd.DataFrame(data["DetData"])


# ----------------------------------------------------------------------
# Dataset functions


def _handleDSArgs(DatasetID, ObsID):
    """Internal function to handle DatasetID/ObsID arguments.

    This function receives the DatasetID and ObsID passed in
    a call, ensures that only one is set, and then converts the value
    into a list if necessary.

    Parameters
    ----------

    DatasetID : int or None
        The DatasetID specified by the user
    ObsID : int or None
        The ObsID specified by the user

    Returns
    -------

    varName : str
        Either 'DatasetID' or 'ObsID', the name of the argument to
            send to the API call.
    varVal : list
        A list of the Dataset/Obs IDs supplied
    single: bool
        Whether a single value, instead of a list, was supplied.

    """
    if (DatasetID is None) == (ObsID is None):
        raise ValueError("Exactly one of `DatasetID` or `ObsID` must be set.")

    varName = None
    varVal = None
    single = False

    if ObsID is not None:
        varName = "ObsID"
        if isinstance(ObsID, (list, tuple)):
            varVal = ObsID
        else:
            varVal = [ObsID]
            single = True

    else:
        varName = "DatasetID"
        if isinstance(DatasetID, (list, tuple)):
            varVal = DatasetID
        else:
            varVal = [DatasetID]
            single = True

    return (varName, varVal, single)


def getDatasetDetails(DatasetID=None, ObsID=None, cat="LSXPS", silent=True, verbose=False):
    """Get the full set of information for an SXPS dataset.

    This obtains all of the information for a given SXPS dataset as
    displayed on the webpages. This duplicates some of the information
    in the Public_Datasets table, but includes other information such
    as details of the objects detected in that dataset etc.

    DatasetID or ObsID can be set - an error will be raised if both
    (or neither) is set - and can either be a single value, or a list.
    If a single value, then this function returns a single dict
    containing the details of the requested dataset. If a list, then
    the return value will be a list of dicts.

    Will add some notes here about superseded stacks later, when stacks
    are done.

    NOTE: May end up with DatasetID being for LSXPS only.

    Parameters
    ----------

    DatasetID : int or list or tuple, optional
        The SXPS dataset ID.

    ObsID : str or int or list or tuple, optional
        The Swift ObsID

    cat : str
        Which SXPS catalogue the source is in.

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    Returns
    -------

    list or dict
        The source info dict, or a list of dicts if a list of sources
        was supplied.

    """
    if verbose:
        silent = False  # noqa

    varName, varVal, single = _handleDSArgs(DatasetID, ObsID)

    # The argument has now been made into a list, so we can go through
    # that list and submit the request to the API.
    # We will store the results in ret
    ret = {}

    sendData = {"whichCat": cat}
    for i in varVal:
        if verbose:
            print(f"Getting data for {varName} = `{i}`")
        sendData[varName] = i
        # We can just accept this result, as the only key we have to have is "OK", and
        # if that wasn't set we will already have hit an error.
        tmp = base.submitAPICall("getSXPSDatasetInfo", sendData, verbose=verbose)
        j = i
        if varName == "ObsID":
            if not single and isinstance(i, int):
                j = f"{i:011d}"
        if "data" in tmp:
            ret[j] = tmp["data"]
            # panda-ise the sources tables?
        else:
            ret[j] = tmp

        if "Sources" in ret[j]:
            ret[j]["Sources"] = pd.DataFrame(ret[j]["Sources"])

        # for b in BAND_NAMES:
        #     k = f"{b}_sources"
        #     if k in ret[j]:
        #         ret[j][k] = pd.DataFrame(ret[j][k])

    if single:
        return ret[varVal[0]]
    else:
        return ret


def saveDatasetImages(
    DatasetID=None,
    ObsID=None,
    cat="LSXPS",
    bands="all",
    types="all",
    getRegions=False,
    destDir="images",
    subDirs=True,
    clobber=False,
    skipErrors=False,
    silent=True,
    verbose=False,
):
    """Get the full set of information for an SXPS dataset.

    This obtains all of the information for a given SXPS dataset as
    displayed on the webpages. This duplicates some of the information
    in the Public_Datasets table, but includes other information such
    as details of the objects detected in that dataset etc.

    DatasetID or ObsID can be set - an error will be raised if both
    (or neither) is set - and can either be a single value, or a list.
    If a single value, then this function returns a single dict
    containing the details of the requested dataset. If a list, then
    the return value will be a list of dicts.

    Will add some notes here about superseded stacks later, when stacks
    are done.

    NOTE: May end up with DatasetID being for LSXPS only.

    Parameters
    ----------

    DatasetID : int or list or tuple, optional
        The SXPS dataset ID.

    ObsID : str or int or list or tuple, optional
        The Swift ObsID

    cat : str
        Which SXPS catalogue the source is in.

    bands : str or list, optional
        Bands (default: 'all').

    types : str or list, optional
        Types (default: 'all').

    getRegions : bool, optional
        Whether to get ds9 region files for the sources in the image
        as well (default: ``False``).

    subDirs : bool, optional
        Whether to save each dataset's data into its own subdirectory
        (default: ``True``).

    clobber : bool
        Whether to overwrite files if they exist.

    skipErrors : bool, optional
        If an error occurs saving a file, do not raise a RuntimeError
        but simply continue to the next file (default ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    Returns
    -------

    list or dict
        The source info dict, or a list of dicts if a list of sources
        was supplied.

    """
    if verbose:
        silent = False

    varName, varVal, single = _handleDSArgs(DatasetID, ObsID)

    # The argument has now been made into a list, so we can go through
    # that list and submit the request to the API.
    # We will store the results in ret

    base._createDir(destDir, silent=silent, verbose=verbose)

    sendData = {"whichCat": cat, "bands": bands, "types": types, "getRegions": getRegions}

    for i in varVal:
        if verbose:
            print(f"Getting data for {varName} = `{i}`")
        sendData[varName] = i
        # We can just accept this result, as the only key we have to have is "OK", and
        # if that wasn't set we will already have hit an error.
        tmp = base.submitAPICall("getSXPSDatasetImages", sendData, verbose=verbose)
        if "SupersededBy" in tmp:
            print("\n ** WARNING: This dataset is an obsolete stack ** \n")
        path = destDir
        if subDirs:
            j = i
            if varName == "ObsID" and isinstance(i, int):
                j = f"{i:011d}"
            path = f"{path}/{j}"
            base._createDir(path, silent=silent, verbose=verbose)

        for b in BAND_NAMES:
            if b in tmp:
                for t in tmp[b]:
                    url = tmp[b][t]
                    if (url == "NOTFOUND") and (not silent):
                        print(f"No {t} image exists for the {b} band for `{i}`.")
                    else:
                        ok = dl._saveURLToFile(url, path, clobber=clobber, silent=silent, verbose=verbose)
                        if not ok:
                            if skipErrors:
                                continue
                            raise RuntimeError(f"Failed to save {url} to {path}/")
        if "Expmap" in tmp:
            url = tmp["Expmap"]
            if (url == "NOTFOUND") and (not silent):
                print(f"No exposure map image exists for `{i}`.")
            else:
                ok = dl._saveURLToFile(url, path, clobber=clobber, silent=silent, verbose=verbose)
                if not ok:
                    if skipErrors:
                        continue
                    raise RuntimeError(f"Failed to save {url} to {path}/")


# ----------------------------------------------------------------------
# Catalogue table functions


def listOldTables(silent=True, verbose=False):
    """Get the list of old tables.

    This returns details of old snapshot versions of the database tables
    that exist.

    Paramters
    ---------

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    Returns
    -------

    dict
        A ``dict`` with keys 'hourly' and 'daily'. These are themselves
        ``dicts``s with keys of epochs for which snapshots exists,
        and those then contain the tables that exist for those
        snapshots.

    """
    if verbose:
        silent = False  # noqa
    sendData = {}
    tmp = base.submitAPICall("getOldSXPSTables", sendData, verbose=verbose, minKeys=("hourly", "daily"))
    return tmp


def getFullTable(
    cat="LSXPS",
    table=None,
    subset=None,
    epoch=None,
    format="csv",
    destDir=None,
    saveData=True,
    returnData=False,
    clobber=False,
    silent=True,
    verbose=False,
):
    """Get the full catalogue data for a given table.

    This allows you to access the entire catalogue in one go. This
    will download the .csv file for the table you requested and read it
    into a pandas ``DataFrame`` which will be returned, and then the
    file will be deleted. You can change this behaviour to keep the file
    and/or to not return a ``DataFrame``.

    Parameters
    ----------

    cat : str, optional
        Which SXPS catalogue to download (default: 'LSXPS').

    table : str
        The database table to download.

    subset : str, optional
        Some tables have subsets you can select, e.g. the 'clean' set of
        Public_Sources. You can specify that subset using this parameter
        or, if this is not supplied, the default subset will be obtained.

    epoch : str, optional
        If supplied, gets an older version of the table from the
        specified epoch (default: ``None``).

    format : str
        The format to download the file. Can be 'csv' or 'fits'.
        NOTE: if ``returnData`` is ``True`` this must be csv (default:
        'csv').

    destDir : str, optional
        The directory to save the file in. If not specified, a temporary
        file will be used.

    saveData : bool, optional
        Whether to keep the csv file (default ``True``).

    returnData : bool, optional
        Whether to return a ``DataFrame`` (default ``False``).

    clobber : bool, optional
        Whether to overwrite the file if it exists. Only matters if
        ``destDir`` is set (default: ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    Returns
    -------

    pandas.DataFrame, optional
        If ``returnData`` was ``True`` then this contains the downloaded
        data.

    """
    if verbose:
        silent = False

    if (table is None) or not isinstance(table, str):
        raise RuntimeError("`table` is a required parameter and must be a string")

    if table.lower() in tableLookup:
        table = tableLookup[table.lower()]

    sendData = {"whichCat": cat, "table": table}

    if subset is not None:
        if not isinstance(subset, str):
            raise RuntimeError("`subset` must be a string")
        sendData["subset"] = subset

    if epoch is not None:
        if not isinstance(epoch, str):
            raise RuntimeError("`epoch` must be a string")
        sendData["epoch"] = epoch

    if not isinstance(format, str):
        raise RuntimeError("`format` must be a string")
    sendData["format"] = format

    if returnData and (format.lower() != "csv"):
        raise RuntimeError("With returnData set, format must be 'csv'")

    tmp = base.submitAPICall("getSXPSTable", sendData, verbose=verbose, minKeys=("URL", "FILE"))
    tmpDir = None
    if destDir is not None:
        base._createDir(destDir, silent=silent, verbose=verbose)
    elif saveData:
        destDir = "."
    else:
        tmpDir = tempfile.TemporaryDirectory()
        destDir = tmpDir.name

    dl._saveURLToFile(
        tmp["URL"],
        destDir,
        prefix=None,
        name=None,
        clobber=clobber,
        silent=silent,
        verbose=verbose,
    )

    ff = os.path.basename(tmp["FILE"])
    fname = f"{destDir}/{ff}"

    # Do we need to unzip it?
    if fname[-3:] == ".gz":
        comm = ["gunzip", fname]
        if verbose:
            print(f"gunzipping {fname}")
        status = subprocess.run(comm, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if status.returncode != 0:
            raise RuntimeError(f"Could not gunzip the downloaded file {fname}")
        fname = fname[:-3]

    ret = None
    if returnData:
        ret = pd.read_csv(fname, sep=",", low_memory=False)

    if not saveData:
        if verbose:
            print(f"Removing downloaded file {fname}")
        os.unlink(fname)
        if tmpDir is not None:
            os.rmdir(tmpDir.name)

    if returnData:
        return ret


# ----------------------------------------------------------------------
# xrt_prods functions


def makeProductRequest(
    email,
    sourceID=None,
    sourceName=None,
    sourceDetails=None,
    getMissingInfo=True,
    cat="LSXPS",
    T0=None,
    useObs="all",
    addProds=None,
    skipErrors=False,
    transient=False,
    transAsSource=False,
    silent=True,
    verbose=False,
):
    """Build XRTProductRequest object(s) for the specified source(s).

    This function will create XRTProductRequest objects for the sources
    supplied in the arguments, either a single XRTProductRequest or a
    list thereof, depending on the supplied arguments. Please read the
    argument definitions with care, since the nature of the returned
    XRTProductRequest is heavily affected by these. Note that the
    XRTProductRequest objects built are NOT submitted, thus you can
    inspect and edit them.

    sourceID or sourceName can be set - an error will be raised if both
    (or neither) is set - and can either be a single value, or a list.
    If it is a single value, then a single XRTProductRequest will be
    returned. If a list, then the return value will be a dict of
    XRTProductRequests, indexed by the sourceID or name, whichever you
    supplied.

    **Note** The parameter ``useObs`` only has any effect if either
    products are specified (via ``addProds``). This is because in an
    ``XRTProductRequest``, the choice of which observations to use is
    set per product, thus this can only be set here if products are
    added at this point; if not, you will have to set the observations
    yourself. The `getObsList()` function may help here.

    **If you create many XRTProductRequest objects, please read the
    documentation about how to throttle submissions, rather than sending
    all job requests en masse.**

    Parameters
    ----------

    email : str
        The email address with which you registered to submit
        XRTProductRequest objects to the servers.

    sourceID :int or list or tuple, optional
        The SXPS source ID.

    sourceName : str or list or tuple, optional
        The SXPS source IAUName.

    sourceDetails : dict, optional
        The data returned by the `getSourceDetails()` function for the
        source(s) requested.

    getMissingInfo : bool, optional
        Whether this function should call `getSourceDetails()` to for any
        sources requested but not present in the `sourceDetails` dict.

    cat : str, optional
        Which SXPS catalogue the source is in. This is only needed if
        you do not supply the ``sourceDetails`` for all requested sources,
        as it will be used to look up the details. (default: 'LSXPS').

    T0 : str or None, optional
        Which value to supply as the T0 parameter of the request
        (default ``None``). Valid values are:

        * ``None``: T0 is not set, but the ``getT0`` global
            parameter will be set to ``True``.

        * 'firstObs': T0 will be set to the start time of the
            first observation of this object in the selected catalogue.

        * 'firstBlindDet': T0 will be set to the start time of
            the first observation in which this object was blindy detected
            in the selected catalogue.

        * 'firstDet': T0 will be set to the start time of the
            first observation in which this object was detected, blindly or
            via a retrospective detection, in the selected catalogue.

    useObs: str, optional
        Which observations should be used to build your requested
        observations (default: 'all').

        Valid values are:

        * 'all': Use all observations in the Swift archive (NB, this may
            include observations not used in the selected catalogue).

        * 'allCat': Use all observations in the selected catalogue
            covering this source region.

        * 'blind': Use only the observations in which this object was
            blindy detected in the selected catalogue.

        * 'allDet': Use only the observations in which this object was
            detected, blindly or via a retrospective detection, in the
            selected catalogue.

    addProds: list, optional
        Whether to add any products to the XRTProductRequest objects
        being created. If this is not ``None`` then it should be a list
        of product types (e.g. 'LightCurve', 'Spectrum' etc.). Any
        product for which the function ``XRTProductRequest.add$prod()``
        exists (where ``$prod`` is the string to include in this list).

    skipErrors : bool, optional
        If an error occurs saving a file, do not raise a RuntimeError
        but simply continue to the next file (default ``False``).

    transient : str, optional
        Whether the object requested is from the LSXPS transient list,
        rather than source list (i.e. the name/ID is a transient name/
        ID) (default: ``False``).

    transAsSource : bool, optional
        Only relevant if ``transient=True``, this means that the object
        identifier supplied is a transient, but the desired light curve
        should be taken from the LSXPS data corresponding to this
        transient. If the transient is not (yet) in LSXPS then this will
        result in an error (default: ``False``).

    silent : bool, optional
        Whether to suppress all output (default: ``True``).

    verbose : bool, optional
        Whether to write verbose output (default: ``False``).

    """
    if verbose:
        silent = False

    # First, handle sourceName or sourceID.
    varName, varVal, single = _handleSourceArgs(sourceID, sourceName)

    # May need to convert transientIDs into sourceIDs
    if transient and transAsSource:
        varVal = _transientToSource(cat, varName, varVal, skipErrors=skipErrors, silent=silent, verbose=verbose)

    # so varName is now either `sourceName` or `sourceID`, and varVal
    # is a list of sourceNames or IDs.
    # single will determine whether are returning a dict or a single thing
    ret = {}
    # Go over the input list:
    for i in varVal:
        if verbose:
            print(f"Building XRTProductRequest for {varName} = `{i}`.")

        # if the sourceDetails is missing, don't want to add it to the sourceDetails
        # dict that the user supplied, as that will pollute that dict.
        si = None
        # check if sourceDetails exists
        if (sourceDetails is None) or (i not in sourceDetails):
            # It may be that the user supplied a sourceDetails string with
            # only one entry, corresponding to this source.
            # So check a random key that is in sourceDetails:
            if single and (sourceDetails is not None) and ("IAUName" in sourceDetails):
                si = sourceDetails
                # Warn??
            else:
                if getMissingInfo:
                    data = {
                        varName: i,
                        "cat": cat,
                        "silent": silent,
                        "verbose": verbose,
                    }
                    if not silent:
                        print(f"Have to get sourceDetails for {varName}={i}")
                    si = None
                    if transient and not transAsSource:
                        si = getTransientDetails(**data)
                    else:
                        si = getSourceDetails(**data)
                elif skipErrors:
                    if not silent:
                        print(f"No entry for {varName}={i} in sourceDetails, SKIPPING.")
                        continue
                else:
                    raise ValueError(f"No entry for {varName}={i} in sourceDetails.")
        else:
            si = sourceDetails[i]
        # Now pass everything to a sub function to actually make the
        # request - that func will either return an XRTProductRequest,
        # or a dict of keys [useObs, request].
        asTrans = transient and not transAsSource
        tmp = _makeSingleProductRequest(
            email,
            si,
            T0=T0,
            useObs=useObs,
            addProds=addProds,
            silent=silent,
            verbose=verbose,
            transient=asTrans,
        )

        if isinstance(tmp, dict) and ("ERROR" in tmp):
            if skipErrors:
                if not silent:
                    print(f"Error: {tmp['ERROR']} for {varName}={i}, SKIPPING.")
                    continue
            else:
                raise RuntimeError(f"Error: {tmp['ERROR']} for {varName}={i}.")

        ret[i] = tmp

    if single:
        return ret[varVal[0]]
    else:
        return ret


def _makeSingleProductRequest(
    email,
    sourceDetails,
    T0=None,
    useObs=None,
    addProds=None,
    transient=False,
    silent=True,
    verbose=False,
):
    """Internal function to actually build an XRTProductRequest."""
    if verbose:
        silent = False

    from ...xrt_prods import XRTProductRequest

    req = XRTProductRequest(email, silent=silent)
    # Now get some global parameters
    gp = {"posErr": 1, "centroid": False, "useSXPS": True}
    # Lookups from sourceDetails pars to XRTProductRequest global pars
    lookUps = {"IAUName": "name", "RA": "RA", "Decl": "Dec"}
    for ipar, rpar in lookUps.items():
        if ipar not in sourceDetails:
            return {"ERROR": f"Cannot find {ipar} in sourceDetails"}

        gp[rpar] = sourceDetails[ipar]
        if ipar in ("RA", "Decl"):
            gp[rpar] = float(gp[rpar])

    T0LookUp = {
        "firstobs": "FirstObsMET",
        "firstblinddet": "FirstBlindDetMET",
        "firstdet": "FirstDetMET",
    }
    if transient:
        T0LookUp = {
            "discovery": "DetectionMET",
        }

    # What to do with T0?
    if T0 is None:
        # T0 is not set, to tell user objects to do a look up.
        gp["getT0"] = True
    elif T0.lower() in T0LookUp:
        # T0 is set to one of the allowed values, so find the corresponding
        # key in sourceDetails, and set it.
        p = T0LookUp[T0.lower()]
        if p not in sourceDetails:
            return {"ERROR": f"Cannot find T0 parameter `{p}` in sourceDetails"}
        else:
            if verbose:
                print(f"Trying to get {p}")
            gp["T0"] = float(sourceDetails[p])
    else:
        return {"ERROR": f"`{T0}` is not a valid T0 value."}

    # Now handle useObs.
    # Do this in two pieces.
    # First, we always have to decided about targetIDs, so do that

    # If useObs is 'all', we need to set getTargs, and if not, then we need
    # to work out the targets ourselves, from the catalogue.
    uoString = None  # The default.
    prodPars = {}
    if useObs == "all":
        gp["getTargs"] = True
        prodPars["whichData"] = "all"
    elif useObs is not None:
        # OK, work out all observations and targetIDs.
        if transient:
            # discovery is handled slighlty differently to the others as it has no target and is a string, not list
            if useObs.lower() == "discovery":
                uoString = sourceDetails["DiscoveryObs"]
                gp["targ"] = uoString[0:8]
            elif useObs.lower() == "new":
                uoString = ",".join(sourceDetails["AllNewObs"])
                gp["targ"] = ",".join(map(str, sourceDetails["AllNewTargs"]))
            else:
                raise RuntimeError("UseObs must be 'all', 'new' or 'discovery'")

        else:
            # Don't use stacks, of course, as these are not obsIDs
            # Use dicts, just to make sure I'm not getting duplicates
            # targList = {}
            # obsList = {}
            # # We always want blind detections
            # z = _getObsList(sourceDetails, "Detections", targList, obsList, False)
            # if z is not None:
            #     return z

            # # Now if callCat/allDet we need to add to this.

            # if useObs == "allCat":
            #     z = _getObsList(sourceDetails, "NonDetections", targList, obsList, False)
            #     if z is not None:
            #         return z
            # elif useObs == "allDet":
            #     z = _getObsList(sourceDetails, "NonDetections", targList, obsList, True)
            #     if z is not None:
            #         return z
            # elif useObs != "blind":
            #     raise ValueError(f"`{useObs}` is not a valid useObs value.")
            tmp = getObsList(sourceDetails, useObs=useObs)

            # Now we have all the observations and targets, so update things
            gp["targ"] = ",".join(tmp["targList"])
            uoString = ",".join(tmp["obsList"])

        prodPars["whichData"] = "user"
        prodPars["useObs"] = uoString

    # Now we can set the global parameters
    req.setGlobalPars(**gp)

    # Finally, set the products, if any.
    # Note: I do not here check if the products are valid; that would mean
    # keeping a separate list of products here, and having to edit this
    # every time we add a new product. If user requests a non-existant
    # product, this will raise an error.

    if addProds is not None:
        for p in addProds:
            req.addProduct(p, clobber=True, **prodPars)

    return req


def getObsList(
    sourceDetails=None, sourceID=None, sourceName=None, cat="LSXPS", useObs=None, silent=True, verbose=False
):
    """Get the list of catalogue observations covering the source"""

    if verbose:
        silent = False

    if sourceDetails is None:
        sourceDetails = getSourceDetails(
            sourceID=sourceID, sourceName=sourceName, cat=cat, silent=silent, verbose=verbose
        )
    else:
        pass

    single = False
    varVal = None
    if "IAUName" in sourceDetails:
        single = True
        varVal = ("dummy",)
        tmp = {"dummy": sourceDetails}
        sourceDetails = tmp
    else:
        single = False
        varVal = sourceDetails.keys()

    ret = {}
    for i in varVal:
        targList = {}
        obsList = {}

        if useObs.lower() not in ("allcat", "alldet", "blind", "allnondet", "retro"):
            raise ValueError(f"{useObs} is not a valid value for `useObs`.")

        # If we want blind detections
        if useObs.lower() in ("allcat", "alldet", "blind"):
            if verbose:
                print("Getting blind detections")
            tmp = _getObsList(sourceDetails[i], "Detections", False)
            targList.update(tmp["targList"])
            obsList.update(tmp["obsList"])

        # If wanted, get all non-detections as well
        if useObs.lower() in ("allcat", "allnondet"):
            if verbose:
                print("Getting non-detections")
            tmp = _getObsList(sourceDetails[i], "NonDetections", False)
            targList.update(tmp["targList"])
            obsList.update(tmp["obsList"])

        # Or if we wanted retrospective detections
        elif useObs.lower() in ("alldet", "retro"):
            if verbose:
                print("Getting retrospective detections")
            tmp = _getObsList(sourceDetails[i], "NonDetections", True)
            targList.update(tmp["targList"])
            obsList.update(tmp["obsList"])

        ret[i] = {"targList": list(targList.keys()), "obsList": list(obsList.keys())}

    if single:
        return ret[varVal[0]]
    else:
        return ret


def _getObsList(sourceDetails, key, detOnly):
    targList = {}
    obsList = {}
    if key not in sourceDetails:
        raise RuntimeError(f"Cannot find {key} in sourceDetails")
    if "Observations" in sourceDetails[key]:
        if "ObsID" not in sourceDetails[key]["Observations"].columns:
            raise RuntimeError("sourceDetails has the wrong format. Is your module out of date?")

        # Fork here on detOnly to do it the quick way if we are getting everything
        if detOnly:
            for i, r in sourceDetails[key]["Observations"].iterrows():
                o = r["ObsID"]
                if isinstance(o, int):
                    o = f"{o:011d}"
                if len(o) != 11:
                    return {"ERROR": "sourceDetails has the wrong ObsID format. Is your module out of date?"}
                isDet = False
                for b in BAND_NAMES:
                    k = f"{b}_Info"
                    if ("isDet" in r[k]) and (int(r[k]["isDet"]) == 1):
                        isDet = True
                        break
                if isDet:
                    obsList[o] = 1
                    targ = o[0:8]
                    targList[targ] = 1

        else:
            for o in sourceDetails[key]["Observations"]["ObsID"]:
                if isinstance(o, int):
                    o = f"{o:011d}"
                if len(o) != 11:
                    return {"ERROR": "sourceDetails has the wrong ObsID format. Is your module out of date?"}
                obsList[o] = 1
                targ = o[0:8]
                targList[targ] = 1
    ret = {"targList": targList, "obsList": obsList}
    return ret



# Lookup table for XRT modes
import re


XRTMODES = {
    0: "Auto",
    1: "Null",
    2: "ShortIM",
    3: "LongIM",
    4: "PUPD",
    5: "LRPD",
    6: "WT",
    7: "PC",
    8: "Raw",
    9: "Bias",
    150: "PC_150",
    200: "PC_200",
    255: "Manual",
    None: "Unset",
}
MODESXRT = {
    "Auto": 0,
    "Null": 1,
    "ShortIM": 2,
    "LongIM": 3,
    "PUPD": 4,
    "LRPD": 5,
    "WT": 6,
    "PC": 7,
    "Raw": 8,
    "Bias": 9,
    "PC_150": 150,
    "PC_200": 200,
    "Manual": 255,
}


class TOOAPI_Instruments:
    """Mixin for XRT / UVOT / BAT mode display and capture"""

    _uvot = None
    _xrt = None
    _bat = None  # For now, bat mode is just an integer

    def uvot_mode_setter(self, attr, mode):
        if type(mode) == str and "0x" in mode:
            """Convert hex string to int"""
            uvot_mode = re.match(r"0x([0-9a-fA-F]+)", mode)
            if uvot_mode is not None:
                setattr(self, f"_{attr}", int(uvot_mode.group(0), 16))
            else:
                setattr(self, f"_{attr}", mode)
            setattr(self, f"_{attr}", int(mode.split(":")[0], 16))
        elif type(mode) == str:
            """Convert decimal string to int"""
            try:
                setattr(self, f"_{attr}", int(mode))
            except (TypeError, ValueError):
                setattr(self, f"_{attr}", mode)
        else:
            """Pass through anything else"""
            setattr(self, f"_{attr}", mode)

    def xrt_mode_setter(self, attr, mode):
        if type(mode) == str:
            if mode in MODESXRT.keys():
                setattr(self, f"_{attr}", MODESXRT[mode])
            else:
                raise NameError(f"Unknown mode ({mode}), should be PC, WT or Auto")
        elif mode is None:
            setattr(self, f"_{attr}", mode)
        else:
            if mode in XRTMODES.keys():
                setattr(self, f"_{attr}", mode)
            else:
                raise ValueError(f"Unknown mode ({mode}), should be PC (7), WT (6) or Auto (0)")

    @property
    def xrt(self):
        """Given a XRT mode number returns a string containing the name of the
        mode"""
        return XRTMODES[self._xrt]

    @xrt.setter
    def xrt(self, mode):
        self.xrt_mode_setter("xrt", mode)

    @property
    def uvot(self):
        """Given a UVOT mode number returns a string containing the name of the
        mode"""
        if type(self._uvot) == int:
            return f"0x{self._uvot:04x}"
        else:
            return self._uvot

    @uvot.setter
    def uvot(self, mode):
        self.uvot_mode_setter("uvot", mode)

    @property
    def bat(self):
        """Given a BAT mode number returns a string containing the name of the
        mode"""
        if type(self._bat) == int:
            return f"0x{self._bat:04x}"
        else:
            return self._bat

    @bat.setter
    def bat(self, mode):
        self.uvot_mode_setter("bat", mode)

    @property
    def uvot_mode_approved(self):
        """Return UVOT as a hex string. Stored as a number internally"""
        if type(self._uvot_mode_approved) == int:
            return f"0x{self._uvot_mode_approved:04x}"
        else:
            return self._uvot_mode_approved

    @uvot_mode_approved.setter
    def uvot_mode_approved(self, mode):
        self.uvot_mode_setter("uvot_mode_approved", mode)

    # Next for approved XRT / UVOT modes from TOORequests
    @property
    def xrt_mode_approved(self):
        """Return XRT mode as abbreviation string. Internally stored as a number."""
        if self._xrt_mode_approved is None:
            return "Unset"
        else:
            return XRTMODES[self._xrt_mode_approved]

    @xrt_mode_approved.setter
    def xrt_mode_approved(self, mode):
        """Allow XRT mode to be set either as a string (e.g. "WT") or as a number (0, 6 or 7)."""
        self.xrt_mode_setter("xrt_mode_approved", mode)

    # Aliases
    xrt_mode = xrt
    uvot_mode = uvot
    bat_mode = bat



import re


class TOOAPI_ObsID:
    """Mixin for handling target ID / Observation ID with various aliases"""

    _target_id = None
    _seg = None

    def convert_obsnum(self, obsnum):
        """Convert various formats for obsnum (SDC and Spacecraft) into one format (Spacecraft)"""
        if type(obsnum) == str:
            if re.match("^[0-9]{11}?$", obsnum) is None:
                raise ValueError("obsnum string format incorrect")
            else:
                targetid = int(obsnum[0:8])
                segment = int(obsnum[8:12])
                return targetid + (segment << 24)
        elif type(obsnum) == int:
            return obsnum
        elif obsnum is None:
            return None
        else:
            raise ValueError("obsnum in wrong format.")

    @property
    def target_id(self):
        return self._target_id

    @target_id.setter
    def target_id(self, tid):
        if type(tid) == str:
            self._target_id = int(tid)
        else:
            self._target_id = tid

    @property
    def seg(self):
        return self._seg

    @seg.setter
    def seg(self, segment):
        self._seg = segment

    @property
    def obsnum(self):
        """Return the obsnum in SDC format"""
        if self._target_id is None or self._seg is None:
            return None
        elif type(self._target_id) == list:
            return [f"{self.target_id[i]:08d}{self.seg[i]:03d}" for i in range(len(self._target_id))]
        else:
            return f"{self.target_id:08d}{self.seg:03d}"

    @obsnum.setter
    def obsnum(self, obsnum):
        """Set the obsnum value, by figuring out what the two formats are."""
        # Deal with lists of obsnumbers
        if type(obsnum) == list and len(obsnum) > 0:
            self._target_id = list()
            self._seg = list()
            for on in obsnum:
                onsc = self.convert_obsnum(on)
                self._target_id.append(onsc & 0xFFFFFF)
                self._seg.append(onsc >> 24)

        elif obsnum is not None and obsnum != []:
            obsnum = self.convert_obsnum(obsnum)
            self._target_id = obsnum & 0xFFFFFF
            self._seg = obsnum >> 24

    @property
    def obsnumsc(self):
        """Return the obsnum in spacecraft format"""
        if type(self._target_id) == list:
            return [self._target_id[i] + (self._seg[i] << 24) for i in range(len(self._target_id))]
        return self._target_id + (self._seg << 24)

    # Aliases
    targetid = target_id
    targid = target_id
    segment = seg
    obsid = obsnum
    obsidsc = obsnumsc



from .api_common import TOOAPI_Baseclass
from .api_daterange import TOOAPI_Daterange
from .api_resolve import TOOAPI_AutoResolve
from .api_skycoord import TOOAPI_SkyCoord
from .api_status import TOOStatus
from .swift_calendar import Swift_Calendar, Swift_Calendar_Entry
from .swift_toorequest import Swift_TOO_Request


class Swift_TOO_Requests(TOOAPI_Baseclass, TOOAPI_Daterange, TOOAPI_SkyCoord, TOOAPI_AutoResolve):
    """Class used to obtain details about previous TOO requests.

    Attributes
    ----------
    username: str
        Swift TOO API username (default 'anonymous')
    shared_secret: str
        TOO API shared secret (default 'anonymous')
    entries : list
        List of TOOs (`Swift_TOO_Request`)
    status : TOOStatus
        Status of API request
    detail : boolean
        Return detailed TOO information (only valid if username matches TOO)
    begin : datetime
        begin time of TOO window
    end : datetime
        end time of TOO window
    length : timedelta
        length of TOO window
    limit : int
        maximum number of TOOs to retrieve
    too_id : int
        ID number of TOO to retrieve
    year : int
        fetch a year of TOOs
    ra : float
        Right Ascension of TOO target in J2000 (decimal degrees)
    dec : float
        Declination of TOO target in J2000 (decimal degrees)
    skycoord : SkyCoord
        SkyCoord version of RA/Dec if astropy is installed
    radius : float
        radius in degrees to search for TOOs
    """

    # Parameters for requested
    _parameters = [
        "username",
        "limit",
        "too_id",
        "year",
        "detail",
        "ra",
        "dec",
        "radius",
        "begin",
        "length",
    ]
    # Local and alias parameters
    _local = ["name", "skycoord", "end", "shared_secret"]
    # Returned values
    _attributes = ["entries", "status"]
    # Returned classes
    _subclasses = [
        Swift_TOO_Request,
        TOOStatus,
        Swift_Calendar,
        Swift_Calendar_Entry,
    ]
    # API name
    api_name = "Swift_TOO_Requests"

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        username: str
            Swift TOO API username (default 'anonymous')
        shared_secret: str
            TOO API shared secret (default 'anonymous')
        detail : boolean
            Return detailed TOO information (only valid if username matches TOO)
        begin : datetime
            begin time of TOO window
        end : datetime
            end time of TOO window
        length : timedelta
            length of TOO window
        limit : int
            maximum number of TOOs to retrieve
        too_id : int
            ID number of TOO to retrieve
        year : int
            fetch a year of TOOs
        ra : float
            Right Ascension of TOO target in J2000 (decimal degrees)
        dec : float
            Declination of TOO target in J2000 (decimal degrees)
        skycoord : SkyCoord
            SkyCoord version of RA/Dec if astropy is installed
        radius : float
            radius in degrees to search for TOOs
        """
        # Parameter values
        self.username = "anonymous"
        self.year = None  # TOOs for a specific year
        # Return detailed information (only returns your TOOs)
        self.detail = None
        # Limit the number of returned TOOs. Default limit is 10.
        self.limit = None
        self.too_id = None  # Request a TOO of a specific TOO ID number
        self.ra = None  # Search on RA / Dec
        self.dec = None  # Default radius is 11.6 arc-minutes
        self.radius = None  # which is the XRT FOV.
        self.skycoord = None  # SkyCoord support in place of RA/Dec
        self.begin = None  # Date range parameters Begin
        self.end = None  # End
        self.length = None  # and length.
        # Request status
        self.status = TOOStatus()  #

        # Results
        self.entries = list()

        # Parse argument keywords
        self._parseargs(*args, **kwargs)

        # See if we pass validation from the constructor, but don't record
        # errors if we don't
        if self.validate():
            self.submit()
        else:
            self.status.clear()

    def __getitem__(self, index):
        return self.entries[index]

    def __len__(self):
        return len(self.entries)

    def by_id(self, too_id):
        """Return Swift_TOO_Request object for a given too_id.

        Parameters
        ----------
        too_id : id
            TOO ID number

        Returns
        -------
        Swift_TOO_Request
            TOO request matching the given too_id
        """
        return {t.too_id: t for t in self.entries}[too_id]

    def validate(self):
        """Validate API submission before submit

        Returns
        -------
        bool
            Was validation successful?
        """
        if (
            self.shared_secret is not None
            and (self.ra is not None and self.dec is not None)
            or self.year is not None
            or self.too_id is not None
            or self.begin is not None
            or self.length is not None
            or self.limit is not None
        ):
            return True
        else:
            return False

    @property
    def _table(self):
        table_cols = [
            "too_id",
            "source_name",
            "instrument",
            "ra",
            "dec",
            "uvot_mode_approved",
            "xrt_mode_approved",
            "timestamp",
            "l_name",
            "urgency",
            "date_begin",
            "date_end",
            "target_id",
        ]
        if len(self.entries) > 0:
            header = [self.entries[0].varnames[col] for col in table_cols]
        else:
            header = []
        t = list()
        for e in self.entries:
            t.append([getattr(e, col) for col in table_cols])
        return header, t


# PEP8 compliant shorthand alias
TOORequests = Swift_TOO_Requests
# Future API compat
Swift_TOORequests = Swift_TOO_Requests



"""Filters for the database queries.

This module defines a single class, 'filter', which is used to compile
filters for database queries -- essentially the building blocks of
SQL 'WHERE' statements.

"""

__docformat__ = "restructedtext en"


class filter:
    """
    A filter, or pair of filters on an SQL column.

    This class stores the components of a query on an SQL column, which
    has one or two parts, combined with AND or OR.

    The components are:

    colName : str The name of the column to filter on.

    filter : str The filter - must match a specific subset.

    val : Union[list,float,int,str] The value associated with the
        filter. OPTIONAL

    combiner : str How to combine the two filter (AND/OR) OPTIONAL

    filter2 : str The second filter - must match a specific subset.

    val2 : Union[list,float,int,str] The value associated with the
        second filter. OPTIONAL


    """

    # ---------------------------------------------------------------
    # Static variables

    # _filters details acceptable values for filter/filter2, and how
    # many arguments each needs.
    _filters = {
        "<": 1,
        ">": 1,
        "=": 1,
        "<=": 1,
        ">=": 1,
        "LIKE": 1,
        "BETWEEN": 2,
        "NOT LIKE": 1,
        "!=": 1,
        "IS NULL": 0,
        "IS NOT NULL": 0,
    }
    # Also need to know the acceptable combiners:
    _combs = ["AND", "OR"]

    _dictKeys = ("colName", "filter", "val", "combiner", "filter2", "val2")

    @staticmethod
    def _emptyDict():
        empty = dict()
        for i in filter._dictKeys:
            empty[i] = None
        return empty

    @staticmethod
    def parseFilterFromList(data):
        i = 0
        tmp = filter._emptyDict()
        if len(data) > len(filter._dictKeys):
            raise ValueError(f"You can only supply {len(filter._dictKeys)} entries")
        for w in data:
            tmp[filter._dictKeys[i]] = w
            i = i + 1
        return tmp

    # ---------------------------------------------------------------
    # Constructor

    def __init__(self, filterDef, metadata=None):
        """Create a filter instance.

        This constructor instantiates the filter class, creating a
        filter to apply to the DB query. It can be created in one of two
        ways.

        1) By passing a dict with the following keys --
            some are optional:

            colName: The name of the column to filter on

            filter: The filter to apply ('<', '>', 'IS NULL' etc)

            val: The value to apply to the filter, if appropriate. e.g.
                if the filter is '<', val may be 3.1234. If the filter
                takes no arguments (IS [NOT] NULL) this is ignored.
                If the filter takes two arguments (BETWEEN) this should
                be a list.

            combiner: OPTIONAL: This can be AND or OR, if this filter
                has to components.

            filter2: As filter, for the second clause.

            val2: As val, for the second clause.

        2) By passing a list, whose values are the above keys, in order,
           e.g. [
                    'foo',
                    '<',
                    3,
                    'OR',
                    'foo',
                    'BETWEEN',
                    [7,10]
           ]

        Parameters
        ----------

        filterDef :  str or list or tuple
            The filter definition

        metadata : pandas.DataFrame, optional
            Table metadata; used to check filter is valid.

        """
        # The acutal filter is stored in this dict, which starts of with
        # all keys set to None.
        self._data = filter._emptyDict()

        # Parse the constructor into an temp dict, so on fail we leave
        # the original alone.
        tmpdict = filter._emptyDict()

        # If a string was passed, we will parse it.
        if isinstance(filterDef, (list, tuple)):
            filterDef = filter.parseFilterFromList(filterDef)
        elif not isinstance(filterDef, dict):
            raise ValueError("filterDef should be a string or dict.")

        # Now work through keys and try to add.

        # column name
        if "colName" not in filterDef:
            raise ValueError("'colName' is a required key.")
        else:
            tmpdict["colName"] = filterDef["colName"]

        # May have to do this twice, if combiner is set.
        suffs = ("", "2")
        for s in suffs:
            # -------
            # filter
            numArgs = 0
            k = f"filter{s}"
            if k not in filterDef:
                raise ValueError("`{k}` is a required key.")
            else:
                tmp = filterDef[k].upper()
                if tmp not in filter._filters:
                    raise ValueError(f"`{tmp}` is not a valid filter.")
                tmpdict[k] = tmp
                numArgs = filter._filters[tmp]
            # -------
            # Value
            v = f"val{s}"
            if numArgs > 0:
                if v not in filterDef:
                    raise ValueError(f"`{v}` keyword, required for filter `{filterDef[k]}`, not supplied.")

                tmpval = filterDef[v]
                # If a single arg is needed, should be number or text
                if numArgs == 1:
                    if not isinstance(tmpval, (float, int, str)):
                        raise ValueError("`{v}` must be a string, int or float.")
                    tmpdict[v] = tmpval
                # if multiple args should be list/tuple and the right length
                elif not isinstance(tmpval, (list, tuple)):
                    raise ValueError(f"`{v}` must be a list/tuple for filter `{filterDef[k]}`.")
                elif len(tmpval) != numArgs:
                    raise ValueError(f"`{v}` should have {numArgs} elements; you supplied {len(tmpval)}.")
                else:
                    tmpdict[v] = tmpval

            # -------
            # combiner
            if s == "2":
                continue
            if ("combiner" not in filterDef) or (filterDef["combiner"] is None):
                break
            c = filterDef["combiner"].upper()
            if c not in filter._combs:
                raise ValueError(f"`{filterDef['combiner']}` is not a valid combiner value.")
            tmpdict["combiner"] = c

        # OK, parsed all keys, if we made it here we had no error.
        self._data = tmpdict

        # Last thing is to check that we are valid, if metadata was supplied
        if (metadata is not None) and (not self.isValid(metadata)):
            tmp = self._data["colName"]
            self._data = None

            raise RuntimeError(f"Cannot create filter, {tmp} is not a valid column.")

    def __repr__(self):
        a = f"A UKSSDC database filter object.\n{self._data}"
        return a

    def __str__(self):
        msg = []
        for i in filter._dictKeys:
            if self._data[i] is None:
                continue
            msg.append(str(self._data[i]))
        return " ".join(msg)

    # ----------------------------------------------------
    # Read-only vars as properties:
    # Results
    @property
    def data(self):
        """The results of the query."""
        return self._data

    def isValid(self, metadata):
        """Check if the filter is valid.

        This receives the metadata for the current catalogue.

        Parameters
        ----------

        metadata : pandas.DataFrame
            The table metadata

        Returns
        -------

        bool
            Whether or not the filter is valid.
        """

        return self._data["colName"] in metadata["ColName"].values



from datetime import timedelta

from .api_common import TOOAPI_Baseclass
from .api_daterange import TOOAPI_Daterange
from .api_resolve import TOOAPI_AutoResolve
from .api_skycoord import TOOAPI_SkyCoord
from .api_status import TOOStatus
from .swift_clock import TOOAPI_ClockCorrect
from .swift_data import TOOAPI_DownloadData
from .swift_instruments import TOOAPI_Instruments
from .swift_obsid import TOOAPI_ObsID


class Swift_AFSTEntry(
    TOOAPI_Baseclass,
    TOOAPI_SkyCoord,
    TOOAPI_ObsID,
    TOOAPI_Instruments,
    TOOAPI_ClockCorrect,
    TOOAPI_DownloadData,
):
    """Class that defines an individual entry in the Swift As-Flown Timeline

    Attributes
    ----------
    begin : datetime
        begin time of observation
    settle : datetime
        settle time of the observation
    end : datetime
        end time of observation
    slewtime : timedelta
        slew time of the observation
    targetid : int
        target ID  of the observation
    seg : int
        segment number of the observation
    xrt : str
        XRT mode of the observation
    uvot : str
        Hex string UVOT mode of the observation
    bat : int
        BAT mode of the observation
    exposure : timedelta
        exposure time of the observation
    ra : float
        Right Ascension of pointing in J2000 (decimal degrees)
    dec : float
        Declination of pointing in J2000 (decimal degrees)
    roll : float
        roll angle of the observation (decimal degrees)
    skycoord : SkyCoord
        SkyCoord version of RA/Dec if astropy is installed
    ra_object : float
        RA of the object that is the target of the pointing
    dec_object : float
        dec of the object that is the target of the pointing
    targname : str
        Target name of the primary target of the observation
    """

    # API core definition
    _attributes = [
        "begin",
        "settle",
        "end",
        "ra",
        "dec",
        "roll",
        "targname",
        "targetid",
        "seg",
        "ra_point",
        "dec_point",
        "xrt",
        "uvot",
        "bat",
        "fom",
        "obstype",
    ]
    _parameters = []
    # Variable names
    _varnames = {
        "begin": "Begin Time",
        "settle": "Settle Time",
        "end": "End Time",
        "ra": "RA(J2000)",
        "dec": "Dec(J200)",
        "roll": "Roll (deg)",
        "targname": "Target Name",
        "targetid": "Target ID",
        "seg": "Segment",
        "ra_object": "Object RA(J2000)",
        "dec_object": "Object Dec(J2000)",
        "xrt": "XRT Mode",
        "uvot": "UVOT Mode",
        "bat": "BAT Mode",
        "fom": "Figure of Merit",
        "obstype": "Observation Type",
        "obsnum": "Observation Number",
        "exposure": "Exposure (s)",
        "slewtime": "Slewtime (s)",
    }
    # API name
    api_name = "Swift_AFST_Entry"

    def __init__(self):
        # For backward compat FIXME API 1.3
        self._varnames["ra_point"] = self._varnames["ra_object"]
        self._varnames["dec_point"] = self._varnames["dec_object"]

        # Attributes
        self.begin = None
        self.settle = None
        self.end = None
        self.ra = None
        self.dec = None
        self.ra_object = None
        self.dec_object = None
        self.roll = None
        self.targname = None
        self.targetid = None
        self.seg = None
        self.status = TOOStatus()
        self._subclasses = [TOOStatus]
        # Instrument config
        self._xrt = None
        self._uvot = None
        # Swift_AFST returns a bunch of stuff we don't care about, so just take the things we do
        self.ignorekeys = True

    @property
    def exposure(self):
        return self.end - self.settle

    @property
    def slewtime(self):
        return self.settle - self.begin

    # The following provides compatibility as we changed ra/dec_point to
    # ra/dec_object. These will go away with a future API update. FIXME API 1.3
    @property
    def ra_point(self):
        return self.ra_object

    @ra_point.setter
    def ra_point(self, ra):
        self.ra_object = ra

    @property
    def dec_point(self):
        return self.dec_object

    @dec_point.setter
    def dec_point(self, dec):
        self.dec_object = dec

    # Compat end

    @property
    def _table(self):
        parameters = ["begin", "end", "targname", "obsnum", "exposure", "slewtime"]
        header = [self._header_title(row) for row in parameters]
        return header, [
            [
                self.begin,
                self.end,
                self.targname,
                self.obsnum,
                self.exposure.seconds,
                self.slewtime.seconds,
            ]
        ]


class Swift_Observation(TOOAPI_Baseclass, TOOAPI_DownloadData):
    """Class to summarize observations taken for given observation ID (obsnum).
    Whereas observations are typically one or more individual snapshot, in TOO
    API speak a `Swift_AFSTEntry`, this class summarizes all snapshots into a
    single begin time, end time. Note that as ra/dec varies between each
    snapshot, only `ra_object`, `dec_object` are given as coordinates.

    Attributes
    ----------
    begin : datetime
        begin time of observation
    settle : datetime
        settle time of the observation
    end : datetime
        end time of observation
    slewtime : timedelta
        slew time of the observation
    targetid : int
        target ID  of the observation
    seg : int
        segment number of the observation
    xrt : str
        XRT mode of the observation
    uvot : str
        Hex string UVOT mode of the observation
    bat : int
        BAT mode of the observation
    exposure : timedelta
        exposure time of the observation
    ra_object : float
        RA of the object that is the target of the pointing
    dec_object : float
        dec of the object that is the target of the pointing
    targname : str
        Target name of the primary target of the observation
    """

    # Core API definitions
    api_name = "Swift_Observation"
    _parameters = [
        "begin",
        "end",
        "targname",
        "targetid",
        "seg",
        "ra_object",
        "dec_object",
        "xrt",
        "uvot",
        "bat",
        "entries",
    ]

    def __init__(self):
        # All the Swift_AFSTEntries for this observation
        TOOAPI_ObsID.__init__(self)
        self.entries = Swift_AFST()

    def __getitem__(self, index):
        return self.entries[index]

    def __len__(self):
        return len(self.entries)

    def append(self, value):
        self.entries.append(value)

    def extend(self, value):
        self.entries.extend(value)

    @property
    def targetid(self):
        return self.entries[0].targetid

    @property
    def seg(self):
        return self.entries[0].seg

    @property
    def obsnum(self):
        return self.entries[0].obsnum

    @property
    def targname(self):
        return self.entries[0].targname

    @property
    def ra_object(self):
        if hasattr(self.entries[0], "ra_object"):
            return self.entries[0].ra_object

    @property
    def dec_object(self):
        if hasattr(self.entries[0], "dec_object"):
            return self.entries[0].dec_object

    @property
    def exposure(self):
        return timedelta(seconds=sum([e.exposure.seconds for e in self.entries]))

    @property
    def slewtime(self):
        return timedelta(seconds=sum([e.slewtime.seconds for e in self.entries]))

    @property
    def begin(self):
        return min([q.begin for q in self.entries])

    @property
    def end(self):
        return max([q.end for q in self.entries])

    @property
    def xrt(self):
        return self.entries[0].xrt

    @property
    def uvot(self):
        return self.entries[0].uvot

    @property
    def bat(self):
        return self.entries[0].bat

    @property
    def snapshots(self):
        return self.entries

    # The following provides compatibility as we changed ra/dec_point to
    # ra/dec_object. These will go away in the next version of the API (1.3).
    @property
    def ra_point(self):
        return self.ra_object

    @ra_point.setter
    def ra_point(self, ra):
        self.ra_object = ra

    @property
    def dec_point(self):
        return self.dec_object

    @dec_point.setter
    def dec_point(self, dec):
        self.dec_object = dec

    # Compat end

    @property
    def _table(self):
        if len(self.entries) > 0:
            header = self.entries[0]._table[0]
        else:
            header = []
        return header, [
            [
                self.begin,
                self.end,
                self.targname,
                self.obsnum,
                self.exposure.seconds,
                self.slewtime.seconds,
            ]
        ]

    # Aliases
    obsid = obsnum
    target_id = targetid
    segment = seg


class Swift_Observations(dict, TOOAPI_Baseclass):
    """Adapted dictionary class for containing observations that mostly is just
    to ensure that data can be displayed in a consistent format. Key is
    typically the Swift Observation ID in SDC format (e.g. '00012345012')."""

    @property
    def _table(self):
        if len(self.values()) > 0:
            header = list(self.values())[0]._table[0]
        else:
            header = []
        return header, [self[obsid]._table[1][0] for obsid in self.keys()]


class Swift_AFST(
    TOOAPI_Baseclass,
    TOOAPI_Daterange,
    TOOAPI_SkyCoord,
    TOOAPI_ObsID,
    TOOAPI_AutoResolve,
    TOOAPI_ClockCorrect,
):
    """Class to fetch Swift As-Flown Science Timeline (AFST) for given
    constraints. Essentially this will return what Swift observed and when, for
    given constraints. Constraints can be for give coordinate (SkyCoord or J2000
    RA/Dec) and radius (in degrees), a given date range, or a given target ID
    (targetid) or Observation ID (obsnum).

    Attributes
    ----------
    begin : datetime
        begin time of visibility window
    end : datetime
        end time of visibility window
    length : timedelta / int
        length of visibility window
    ra : float
        Right Ascension of target in J2000 (decimal degrees)
    dec : float
        Declination of target in J2000 (decimal degrees)
    radius : float
        Search radius in degrees
    skycoord : SkyCoord
        SkyCoord version of RA/Dec if astropy is installed
    username: str
        Swift TOO API username (default 'anonymous')
    shared_secret: str
        TOO API shared secret (default 'anonymous')
    entries : list
        List of observations (`Swift_AFSTEntry`)
    status : TOOStatus
        Status of API request
    afstmax: datetime
        When is the AFST valid up to
    """

    # Define API name
    api_name = "Swift_AFST"
    # Contents of the _parameters
    _parameters = [
        "username",
        "begin",
        "end",
        "ra",
        "dec",
        "radius",
        "targetid",
        "obsnum",
    ]
    _local = ["obsid", "name", "skycoord", "length", "target_id", "shared_secret"]
    _attributes = ["status", "afstmax", "entries"]
    # Acceptable classes that be part of this class
    _subclasses = [Swift_AFSTEntry, TOOStatus]

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        begin : datetime
            begin time of window
        end : datetime
            end time of window
        length : timedelta
            length of window
        ra : float
            Right Ascension of target in J2000 (decimal degrees)
        dec : float
            Declination of target in J2000 (decimal degrees)
        skycoord : SkyCoord
            SkyCoord version of RA/Dec if astropy is installed
        radius : float
            radius in degrees to search around (default 0.197)
        targetid : int
            target ID of target
        obsid : int / str
            Observation ID of target, either in spacecraft (int) or SDC (str)
            formats
        username : str
            username for TOO API (default 'anonymous')
        shared_secret : str
            shared secret for TOO API (default 'anonymous')
        """

        # Coordinate search
        self.ra = None
        self.dec = None
        self.radius = 11.8 / 60  # Default 11.8 arcmin - XRT FOV

        # Search on targetid/obsnum
        self.targetid = None
        self.obsnum = None

        # Login
        self.username = "anonymous"
        # AFST entries go here
        self.entries = list()
        # Status of request
        self.status = TOOStatus()

        # Parse argument keywords
        self._parseargs(*args, **kwargs)

        # AFST maximum date
        self.afstmax = None

        # Observations
        self._observations = Swift_Observations()

        # See if we pass validation from the constructor, but don't record
        # errors if we don't
        if self.validate():
            self.submit()
        else:
            self.status.clear()

    @property
    def _table(self):
        if len(self.entries) > 0:
            header = self.entries[0]._table[0]
        else:
            header = []
        return header, [ppt._table[1][0] for ppt in self]

    @property
    def observations(self):
        if len(self.entries) > 0 and len(self._observations.keys()) == 0:
            for q in self.entries:
                self._observations[q.obsnum] = Swift_Observation()
            _ = [self._observations[q.obsnum].append(q) for q in self.entries]
        return self._observations

    def __getitem__(self, index):
        return self.entries[index]

    def __len__(self):
        return len(self.entries)

    def append(self, value):
        self.entries.append(value)

    def validate(self):
        """Make sure that all parameters required for a valid request are
        passed"""
        # How many search keys? Require at least one
        keys = self.api_data.keys()

        # We need at least one of these keys to be submitted
        req_keys = ["begin", "ra", "dec", "targetid", "obsnum"]

        # Check how many of them are in the request
        total_keys = 0
        for key in keys:
            if key in req_keys:
                if self.api_data[key]:
                    total_keys += 1

        # We need at least one key to be set
        if total_keys == 0:
            self.status.error("Please supply search parameters to narrow search.")
            return False

        # Check if ra or dec are in keys, we have both.
        if "ra" in keys or "dec" in keys:
            if not ("ra" in keys and "dec" in keys):
                self.status.error("Must supply both RA and Dec.")
                return False

        return True


# Alias names for class for better PEP8 and future compat
Swift_ObsQuery = Swift_AFST
ObsQuery = Swift_AFST
AFST = Swift_AFST
Swift_AFST_Entry = Swift_AFSTEntry
ObsEntry = Swift_AFSTEntry
Swift_ObsEntry = Swift_AFSTEntry
AFSTEntry = Swift_AFSTEntry



from .api_common import TOOAPI_Baseclass
from .api_daterange import TOOAPI_Daterange, TOOAPI_TriggerTime
from .api_resolve import TOOAPI_AutoResolve
from .api_skycoord import TOOAPI_SkyCoord
from .api_status import TOOStatus
from .swift_calendar import Swift_Calendar
from .swift_instruments import TOOAPI_Instruments
from .swift_obsid import TOOAPI_ObsID

HAS_ASTROPY = False
try:
    import astropy.units as u

    HAS_ASTROPY = True
except ImportError:
    pass


class Swift_TOORequest(
    TOOAPI_Baseclass,
    TOOAPI_SkyCoord,
    TOOAPI_ObsID,
    TOOAPI_Instruments,
    TOOAPI_Daterange,
    TOOAPI_AutoResolve,
    TOOAPI_TriggerTime,
):
    """Class to construct a TOO for submission to Swift MOC. Class provides
    internal validation of TOO, based on simple criteria. Submission is handled
    by creating an signed JSON file, using "shared secret" to ensure that the
    TOO is from the registered party, and uploading via a HTTP POST to the Swift
    website. Verification of the success or failure of submission is reported
    into the TOOStatus class, which is populated using parameters
    reported by the  Swift TOO website upon submission.

    Attributes
    ----------
    l_name : str
        Proposers last name (default None)
    date_begin : datetime
        Date observations are scheduled to begin (default None)
    date_end : datetime
        Date observations are scheduled to end (default None)
    decision : str
        Decision on TOO (default None)
    xrt_mode_approved : int
        Approved XRT mode (default None)
    uvot_mode_approved : int
        Approved UVOT mode (default None)
    exp_time_per_visit_approved : int
        Exposure time per visit approved (default None)
    num_of_visits_approved : int
        Number of visits approved (default None)
    total_exp_time_approved : int
        Total approved exposure time in seconds (default None)
    target_id : int
        Target ID (default None)
    done : boolean
        Is the TOO considered to be complete (default None)
    status : TOOStatus
        status of request
    """

    # Name the class
    api_name = "Swift_TOO_Request"
    # Paramaters that get submitted as part of the JSON
    _parameters = [
        "username",
        "source_name",
        "source_type",
        "ra",
        "dec",
        "poserr",
        "instrument",
        "urgency",
        "opt_mag",
        "opt_filt",
        "xrt_countrate",
        "bat_countrate",
        "other_brightness",
        "grb_detector",
        "immediate_objective",
        "science_just",
        "exposure",
        "exp_time_just",
        "exp_time_per_visit",
        "num_of_visits",
        "monitoring_freq",
        "proposal",
        "proposal_id",
        "proposal_trigger_just",
        "proposal_pi",
        "xrt_mode",
        "uvot_mode",
        "uvot_just",
        "slew_in_place",
        "tiling",
        "number_of_tiles",
        "exposure_time_per_tile",
        "tiling_justification",
        "obs_n",
        "obs_type",
        "grb_triggertime",
        "debug",
        "validate_only",
        "quiet",
    ]
    # Alias parameter names
    _local = ["skycoord", "shared_secret", "xrt", "uvot"]
    # Attributes that may get returned
    _attributes = [
        "too_id",
        "timestamp",
        "decision",
        "done",
        "target_id",
        "date_begin",
        "date_end",
        "l_name",
        "xrt_mode_approved",
        "uvot_mode_approved",
        "calendar",
        "total_exp_time_approved",
        "num_of_visits_approved",
        "exp_time_per_visit_approved",
    ]
    # The three instruments on Swift
    instruments = ["XRT", "BAT", "UVOT"]
    # Common missions that that trigger detections
    mission_names = [
        "Fermi/LAT",
        "Swift/BAT",
        "INTEGRAL",
        "MAXI",
        "IPN",
        "Fermi/GBM",
        "IceCube",
        "LVC",
        "ANTARES",
        "ZTF",
        "ASAS-SN",
    ]
    # Valid units for monitoring frequency. Can add a "s" to each one if you like, so "3 orbits" is good.
    monitoring_units = [
        "second",
        "minute",
        "hour",
        "day",
        "week",
        "month",
        "year",
        "orbit",
    ]

    # English Descriptions of all the variables
    varnames = {
        "decision": "Decision",
        "done": "Done",
        "date_begin": "Begin date",
        "date_end": "End date",
        "calendar": "Calendar",
        "slew_in_place": "Slew in Place",
        "grb_triggertime": "GRB Trigger Time (UT)",
        "exp_time_per_visit_approved": "Exposure Time per Visit (s)",
        "total_exp_time_approved": "Total Exposure (s)",
        "num_of_visits_approved": "Number of Visits",
        "l_name": "Requester",
        "username": "Requester",
        "too_id": "ToO ID",
        "timestamp": "Time Submitted",
        "target_id": "Primary Target ID",
        "sourceinfo": "Object Information",
        "ra": "Right Ascenscion (J2000)",
        "dec": "Declination (J2000)",
        "source_name": "Object Name",
        "resolve": "Resolve coordinates",
        "position_err": "Position Error",
        "poserr": "Position Error (90% confidence - arcminutes)",
        "obs_type": "What is Driving the Exposure Time?",
        "source_type": "Type or Classification",
        "tiling": "Tiling",
        "immediate_objective": "Immediate Objective",
        "proposal": "GI Program",
        "proposal_details": "GI Proposal Details",
        "instrument": "Instrument",
        "tiling_type": "Tiling Type",
        "number_of_tiles": "Number of Tiles",
        "exposure_time_per_tile": "Exposure Time per Tile",
        "tiling_justification": "Tiling Justification",
        "instruments": "Instrument Most Critical to your Science Goals",
        "urgency": "Urgency",
        "proposal_id": "GI Proposal ID",
        "proposal_pi": "GI Proposal PI",
        "proposal_trigger_just": "GI Trigger Justification",
        "source_brightness": "Object Brightness",
        "opt_mag": "Optical Magnitude",
        "opt_filt": "Optical Filter",
        "xrt_countrate": "XRT Estimated Rate (c/s)",
        "bat_countrate": "BAT Countrate (c/s)",
        "other_brightness": "Other Brightness",
        "science_just": "Science Justification",
        "monitoring": "Observation Campaign",
        "obs_n": "Observation Strategy",
        "num_of_visits": "Number of Visits",
        "exp_time_per_visit": "Exposure Time per Visit (seconds)",
        "monitoring_freq": "Monitoring Cadence",
        "monitoring_freq_approved": "Monitoring Cadence",
        "monitoring_details": "Monitoring Details",
        "exposure": "Exposure Time (seconds)",
        "exp_time_just": "Exposure Time Justification",
        "xrt_mode": "XRT Mode",
        "xrt_mode_approved": "XRT Mode (Approved)",
        "uvot_mode": "UVOT Mode",
        "uvot_mode_approved": "UVOT Mode (Approved)",
        "uvot_just": "UVOT Mode Justification",
        "trigger_date": "GRB Trigger Date (YYYY/MM/DD)",
        "trigger_time": "GRB Trigger Time (HH:MM:SS)",
        "grb_detector": "GRB Discovery Instrument",
        "grbinfo": "GRB Details",
        "debug": "Debug mode",
        "validate_only": "Validate only",
        "quiet": "Quiet mode",
    }

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        username : string
            Swift TOO username (default None)
        too_id : int
            TOO ID assigned by server on acceptance (default None)
        timestamp : datetime
            Timestamp that TOO was accepted (default None)
        source_name : string
            Name of the object we're requesting a TOO for (default None)
        source_type : string
            Type of object (e.g. "Supernova", "LMXB", "BL Lac")  (default None)
        ra : float
            RA(J2000) Degrees decimal (default None)
        dec : float
            declination (J2000) Degrees decimal (default None)
        poserr : float
            Position error in arc-minutes (default 0.0)
        instrument : string
            Choices "XRT","UVOT","BAT" (default "XRT")
        urgency : int
            TOO Urgency 1 = Within 4 hours, 2 = within 24 hours, 3 = Days to a
            week, 4 = week - month. (default 3)
        obs_type : str
            Select from obs_types one of four options, e.g. obs_types[1] ==
            'Light Curve'
        opt_mag : float
            UVOT optical magnitude (default None)
        opt_filt : string
            What filter was this measured in (can be non-UVOT filters) (default
            None)
        xrt_countrate : float
            XRT estimated counts/s (default None)
        bat_countrate : float
            BAT estimated counts/s (default None)
        other_brightness : float
            Any other brightness info (default None)
        grb_detector : str
            What detected this GRB (or other) trigger? Should be
            "Mission/Detector" (e.g "Swift/BAT, Fermi/LAT")
        grb_triggertime : datetime
            GRB trigger date/time (default None)
        immediate_objective : string
            One sentence explaination of TOO (default None)
        science_just : string
            this is the Science Justification (default None)
        exposure : int
            this is the user requested exposure in seconds
        exp_time_just : string
            Justification of monitoring exposure (default None)
        exp_time_per_visit : integer
            Exposure per visit in seconds (default None)
        num_of_visits : integer
            Number of visits (default 1)
        monitoring_freq : string / astropy.units
            Formatted text to describe monitoring cadence. E.g. "2 days", "3
            orbits", "1 week". See monitoring_units for valid units (default
            None)
        proposal : boolean
            Is this a GI proposal? (default False)
        proposal_id : string or integer
            What is the GI proposal ID (default None)
        proposal_trigger_just : string
            Note this is the GI Program Trigger Criteria Justification (default
            None)
        proposal_pi : string
            Proposal PI name (default None)
        xrt_mode : int
            XRT instrument mode. 7 = Photon Counting, 6 = Windowed Timing, 0 =
            Auto (self select). PC is default. (default 7)
        uvot_mode : int
            Hex mode for requested UVOT filter. Default FOTD. See Cheat Sheet or
            https://www.swift.psu.edu/operations/mode_lookup.php for codes.
            (default 0x9999)
        uvot_just : str
            Text justification of UVOT filter (default None)
        slew_in_place : boolean
            Perform a slew-in-place? Typically used for GRISM observation.
            Allows the source to be placed more accurately on the detector.
            (default None)
        tiling : boolean
            Is this a tiling request (default False)
        number_of_tiles : int
            Set this if you want a fixed number of tiles. Traditional tiling
            patterns are 4,7,19,37 "circular" tilings. If you don't set this
            we'll calculate based on error radius. (default None)
        exposure_time_per_tile : int
            exposure time per tile in seconds, otherwise it'll just be exposure
            / number_of_tiles (default None)
        tiling_justification : str
            Text description of why tiling is justified (default None)
        debug : boolean
            Debug mode (default None)
        """
        # User chooseable values
        self.username = None  # Swift TOO username (string)
        # Swift TOO shared secret. Log in to Swift TOO page to find out / change your shared secret
        self.shared_secret = None

        # These next two are assigned by the server, so don't set them
        self.too_id = None  # TOO ID assigned by server on acceptance (int)
        self.timestamp = None  # Timestamp that TOO was accepted (datetime)
        # Source name, type, location, position_error
        self.source_name = None  # Name of the object we're requesting a TOO for (string)
        # Type of object (e.g. "Supernova", "LMXB", "BL Lac")  (string)
        self.source_type = None
        self.ra = None  # RA(J2000) Degrees decimal (float)
        self.dec = None  # declination (J2000) Degrees decimal (float)
        self.poserr = 0.0  # Position error in arc-minutes (float)

        # Request details
        self.instrument = "XRT"  # Choices "XRT","UVOT","BAT" (string)
        # 1 = Within 4 hours, 2 = within 24 hours, 3 = Days to a week, 4 = week - month. (int)
        self.urgency = 3

        # Observation Type - primary goal of observation
        self.obs_types = ["Spectroscopy", "Light Curve", "Position", "Timing"]
        # Select from self.obs_types one of four options, e.g. self.obs_types[1] == 'Light Curve'
        self.obs_type = None

        # Description of the source brightness for various instruments
        self.opt_mag = None  # UVOT optical magnitude (float)
        # What filter was this measured in (can be non-UVOT filters) (string)
        self.opt_filt = None

        self.xrt_countrate = None  # XRT estimated counts/s (float)
        self.bat_countrate = None  # BAT estimated counts/s (float)
        self.other_brightness = None  # Any other brightness info (float)

        # GRB stuff
        # Should be "Mission/Detection" (e.g "Swift/BAT, Fermi/LAT") (string)
        self.grb_detector = None
        self.grb_triggertime = None  # GRB trigger date/time (datetime)

        # Science Justification
        # One sentence explaination of TOO (string)
        self.immediate_objective = None
        # Note this is the Science Justification (string)
        self.science_just = None

        # Exposure requested time (total)
        self._exposure = None  # Note this is the user requested exposure

        # Monitoring request
        # Justification of monitoring exposure (string)
        self.exp_time_just = None
        # Exposure per visit in seconds (integer)
        self.exp_time_per_visit = None
        self.num_of_visits = 1  # Number of visits (integer)
        # Formatted text to describe monitoring cadence. E.g. "2 days", "3 orbits", "1 week". See self.monitoring_units for valid units (string)
        self.monitoring_freq = None
        # GI stuff
        self.proposal = None  # Is this a GI proposal? (boolean)
        # What is the GI proposal ID (string or integer)
        self.proposal_id = None
        # Note this is the GI Program Trigger Criteria Justification (string)
        self.proposal_trigger_just = None
        self.proposal_pi = None  # Proposal PI name (string)

        # Instrument mode
        # 7 = Photon Counting, 6 = Windowed Timing, 0 = Auto (self select). PC is default (int)
        self._xrt = 7
        # Hex mode for requested UVOT filter. Default FOTD. See Cheat Sheet or https://www.swift.psu.edu/operations/mode_lookup.php for codes. (int)
        self._uvot = 0x9999
        self.uvot_just = None  # Text justification of UVOT filter (str)
        # Perform a slew-in-place? Typically used for GRISM observation. Allows the source to be placed more accurately on the detector. (boolean)
        self.slew_in_place = None

        # Tiling request
        self.tiling = None  # Is this a tiling request (boolean)
        # Set this if you want a fixed number of tiles. Traditional tiling
        # patterns are 4,7,19,37 "circular" tilings. If you don't set this we'll
        # calculate based on error radius. (int)
        self.number_of_tiles = None
        # Set this if you want to have a fixed tile exposure, otherwise it'll just be exposure / number_of_tiles (int)
        self.exposure_time_per_tile = None
        # Text description of why tiling is justified (str)
        self.tiling_justification = None

        # Calendar
        self._calendar = None

        # More parameters that are assigned server side
        self.l_name = None  # Proposers last name (str)
        # Date observations are scheduled to begin (datetime)
        self.date_begin = None
        # Date observations are scheduled to end (datetime)
        self.date_end = None
        self.decision = None  # Decision on TOO (str)
        self._xrt_mode_approved = None  # Approved XRT mode (int)
        self._uvot_mode_approved = None  # Approved UVOT mode (int)
        # Exposure time per visit approved (int)
        self.exp_time_per_visit_approved = None
        self.num_of_visits_approved = None  # Number of visits approved (int)
        # Total approved exposure time in seconds (int)
        self.total_exp_time_approved = None
        self.target_id = None  # Target ID (int)
        self.done = None  # Is the TOO considered to be complete (boolean)

        # Debug parameter - if this is set, sending a TOO won't actually submit it. Good for testing.
        self.debug = None  # Debug mode (boolean)
        self.quiet = False

        # Status of request
        self.status = TOOStatus()

        # Do a server side validation instead of submit?
        self.validate_only = None

        # Things that can be a subclass of this class
        self._subclasses = [Swift_Calendar]
        self.ignorekeys = True

        # Parse argument keywords
        self._parseargs(*args, **kwargs)

    @property
    def calendar(self):
        """If no calendar data exists for this TOO, fetch it."""
        if self.too_id is not None:
            if self._calendar.too_id is None:
                self._calendar.too_id = self.too_id
                self._calendar.submit()
        return self._calendar

    @calendar.setter
    def calendar(self, cal):
        self._calendar = cal

    @property
    def obs_n(self):
        """Is this a request for a single observation or multiple?"""
        if self.num_of_visits != "" and self.num_of_visits > 1:
            return "multiple"
        else:
            return "single"

    @obs_n.setter
    def obs_n(self, value):
        """Just ignore attempts to set this property"""
        pass

    def validate(self):
        """Validate API submission before submit

        Returns
        -------
        bool
            Was validation successful?
        """
        # Required arguments for a valid TOO
        requirements = [
            "ra",
            "dec",
            "num_of_visits",
            "exp_time_just",
            "source_type",
            "source_name",
            "science_just",
            "username",
            "obs_type",
        ]

        # Clear the status
        self.status.clear()

        # Check that the minimum requirements are met
        for req in requirements:
            if self.__getattribute__(req) is None:
                self.status.error(f"Missing key: {req}")
                return False

        if self.obs_type not in self.obs_types:
            self.status.error(f"Observation Type needs to be one of the following: {self.obs_types}")
            return False

        if self.instrument not in self.instruments:
            self.status.error(f"Instrument name ({self.instrument}) not valid")
            return False

        # If this is monitoring, we need time between exposures
        if self.num_of_visits > 1:
            if self.monitoring_freq is None or self.monitoring_freq == "":
                self.status.error("Need monitoring cadence.")
                return False

            if not self.exp_time_per_visit:
                self.exp_time_per_visit = int(self.exposure / self.num_of_visits)
            else:
                if self.exp_time_per_visit * self.num_of_visits != self.exposure:
                    self.status.warning("INFO: Total exposure time does not match total of individuals. Corrected.")
                    self.exposure = self.exp_time_per_visit * self.num_of_visits
        else:
            if not self.exposure:
                self.exposure = self.exp_time_per_visit

        # Check monitoring frequency is correct
        if self.monitoring_freq is not None:
            if HAS_ASTROPY and type(self.monitoring_freq) is u.quantity.Quantity:
                if self.monitoring_freq.to(u.day).value >= (1 * u.day).value:
                    self.monitoring_freq = f"{self.monitoring_freq.to(u.day).value} days"
                else:
                    self.monitoring_freq = f"{self.monitoring_freq.to(u.hour).value} hours"

            _, unit = self.monitoring_freq.strip().split()
            if unit[-1] == "s":
                unit = unit[0:-1]
            if unit not in self.monitoring_units:
                self.status.error(f"Monitoring unit ({unit}) not valid")
                return False

        # Check validity of GI requests
        gi_requirements = ["proposal_id", "proposal_pi", "proposal_trigger_just"]
        if self.proposal:
            for req in gi_requirements:
                if getattr(self, req) is None:
                    self.status.error(f"Missing key: {req}")
                    return False

        # Check trigger requirements
        grb_requirements = ["grb_triggertime", "grb_detector"]
        if self.source_type == "GRB":
            for req in grb_requirements:
                if getattr(self, req) is None:
                    self.status.error(f"Missing key: {req}")
                    return False

        return True

    def server_validate(self):
        """Validate the TOO request with the TOO API server."""
        # Perform a _local validation first
        self.validate()
        # Do a server side validation
        if len(self.status.errors) == 0:
            # Preserve existing warnings
            warnings = self.status.warnings
            self.validate_only = True
            self.submit()
            self.validate_only = False
            self.status.warnings += warnings
            if len(self.status.errors) == 0:
                return True
            else:
                return False
        else:
            return False

    @property
    def _table(self):
        tab = list()
        if self.decision is not None:
            _parameters = [
                "too_id",
                "l_name",
                "timestamp",
                "urgency",
                "source_name",
                "source_type",
                "grb_triggertime",
                "grb_detector",
                "ra",
                "dec",
                "proposal_pi",
                "proposal_id",
                "proposal_trigger_just",
                "poserr",
                "science_just",
                "opt_mag",
                "opt_filt",
                "xrt_countrate",
                "other_brightness",
                "bat_countrate",
                "exp_time_per_visit_approved",
                "num_of_visits_approved",
                "total_exp_time_approved",
                "monitoring_freq",
                "immediate_objective",
                "exp_time_just",
                "instrument",
                "obs_type",
                "xrt_mode_approved",
                "uvot_mode_approved",
                "uvot_just",
                "target_id",
            ]
        else:
            _parameters = self._parameters
        for row in _parameters:
            val = getattr(self, row)
            if val is not None and val != "":
                tab.append([self.varnames[row], val])
        if len(tab) > 0:
            header = ["Parameter", "Value"]
        else:
            header = []
        return header, tab


# Aliases for class
Swift_TOO = Swift_TOORequest
TOO = Swift_TOORequest
TOORequest = Swift_TOORequest
Swift_TOO_Request = Swift_TOORequest



from tabulate import tabulate

from .api_common import TOOAPI_Baseclass
from .api_resolve import TOOAPI_AutoResolve
from .api_skycoord import TOOAPI_SkyCoord
from .api_status import TOOStatus
from .swift_instruments import TOOAPI_Instruments


class Swift_UVOTModeEntry(TOOAPI_Baseclass):
    """Class describing a single entry in the UVOT Mode table

    Attributes
    ----------
    uvotmode : int
        UVOT mode
    filter_num : int
        filter number
    min_exposure : int
        minumum exposure for entry
    filter_name : str
        filter name
    filter_pos : int
        position of filter in filter wheel
    filter_seqid : int
        Sequence ID of filter
    eventmode : boolean
        Is filter entry taken in event mode
    field_of_view : int
        Filter of view in arcminutes
    binning : int
        Binning of entry
    max_exposure : int
        maximum exposure time for filter
    weight: boolean
        Is observation for this filter weighted for the total exposure time
    special: str
        comment on special modes
    """

    # Core API definitions
    _parameters = [
        "uvotmode",
        "filter_num",
        "min_exposure",
        "filter_name",
        "filter_pos",
        "filter_seqid",
        "eventmode",
        "field_of_view",
        "binning",
        "max_exposure",
        "weight",
        "special",
        "comment",
    ]
    _attributes = []
    api_name = "UVOT_mode_entry"

    def __init__(self):
        # Lazy variable init
        for row in self._parameters:
            setattr(self, row, None)

    def __str__(self):
        return self.filter_name


class Swift_UVOTMode(TOOAPI_Baseclass, TOOAPI_Instruments, TOOAPI_SkyCoord, TOOAPI_AutoResolve):
    """Class to fetch information about a given UVOT mode. Specifically this is
    useful for understanding for a given UVOT hex mode (e.g. 0x30ed), which
    filters and configuration are used by UVOT.

    Attributes
    ----------
    uvotmode : int / str
        UVOT mode to fetch information about (can be hex string or integer)
    username : str
        username for TOO API (default 'anonymous')
    shared_secret : str
        shared secret for TOO API (default 'anonymous')
    status : TOOStatus
        TOO API submission status
    entries : list
        entries (`UVOT_mode_entry`) in UVOT mode table
    """

    # Core API definitions
    _parameters = ["username", "uvotmode", "ra", "dec"]
    _attributes = ["status", "entries"]
    # Local parameters
    _local = ["shared_secret", "name"]
    _subclasses = [Swift_UVOTModeEntry, TOOStatus]
    api_name = "UVOT_mode"
    # Alias for uvotmode
    uvotmode = TOOAPI_Instruments.uvot

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        uvotmode : int / str
            UVOT mode to fetch information about (can be hex string or integer)
        username : str
            username for TOO API (default 'anonymous')
        shared_secret : str
            shared secret for TOO API (default 'anonymous')
        """
        TOOAPI_SkyCoord.__init__(self)
        TOOAPI_AutoResolve.__init__(self)
        # Set up username
        self.username = "anonymous"
        # Set up uvotmode
        self.uvotmode = None
        # Set up coordinates
        self.ra = None
        self.dec = None
        # Here is where the data go
        self.entries = None
        # TOO API status
        self.status = TOOStatus()
        # Parse argument keywords
        self._parseargs(*args, **kwargs)

        # See if we pass validation from the constructor, but don't record
        # errors if we don't
        if self.validate():
            self.submit()
        else:
            self.status.clear()

    def __getitem__(self, index):
        return self.entries[index]

    def __len__(self):
        return len(self.entries)

    def __str__(self):
        """Display UVOT mode table"""
        if hasattr(self, "status") and self.status == "Rejected" and self.status.__class__.__name__ == "TOOStatus":
            return "Rejected with the following error(s): " + " ".join(self.status.errors)
        elif self.entries is not None:
            table_cols = [
                "filter_name",
                "eventmode",
                "field_of_view",
                "binning",
                "max_exposure",
                "weight",
                "comment",
            ]
            table_columns = list()
            table_columns.append(
                [
                    "Filter",
                    "Event FOV",
                    "Image FOV",
                    "Bin Size",
                    "Max. Exp. Time",
                    "Weighting",
                    "Comments",
                ]
            )
            for entry in self.entries:
                table_columns.append([getattr(entry, col) for col in table_cols])

            table = f"UVOT Mode: {self.uvotmode}\n"
            table += "The following table summarizes this mode, ordered by the filter sequence:\n"
            table += tabulate(table_columns, tablefmt="pretty")
            table += "\nFilter: The particular filter in the sequence.\n"
            table += "Event FOV: The size of the FOV (in arc-minutes) for UVOT event data.\n"
            table += "Image FOV: The size of the FOV (in arc-minutes) for UVOT image data.\n"
            table += "Max. Exp. Time: The maximum amount of time the snapshot will spend on the particular filter in the sequence.\n"
            table += "Weighting: Ratio of time spent on the particular filter in the sequence.\n"
            table += "Comments: Additional notes that may be useful to know.\n"
            return table
        else:
            return "No data"

    def _repr_html_(self):
        """Jupyter Notebook friendly display of UVOT mode table"""
        if hasattr(self, "status") and self.status == "Rejected" and self.status.__class__.__name__ == "TOOStatus":
            return "<b>Rejected with the following error(s): </b>" + " ".join(self.status.errors)
        elif self.entries is not None:
            html = f"<h2>UVOT Mode: {self.uvotmode}</h2>"
            html += "<p>The following table summarizes this mode, ordered by the filter sequence:</p>"

            html += '<table id="modelist" cellpadding=4 cellspacing=0>'
            html += "<tr>"  # style="background-color:#08f; color:#fff;">'
            html += "<th>Filter</th>"
            html += "<th>Event FOV</th>"
            html += "<th>Image FOV</th>"
            html += "<th>Bin Size</th>"
            html += "<th>Max. Exp. Time</th>"
            html += "<th>Weighting</th>"
            html += "<th>Comments</th>"
            html += "</tr>"

            table_cols = [
                "filter_name",
                "eventmode",
                "field_of_view",
                "binning",
                "max_exposure",
                "weight",
                "comment",
            ]
            i = 0
            for entry in self.entries:
                if i % 2:
                    html += '<tr style="background-color:#eee;">'
                else:
                    html += '<tr">'
                for col in table_cols:
                    html += "<td>"
                    html += f"{getattr(entry, col)}"
                    html += "</td>"

                html += "</tr>"
            html += "</table>"
            html += '<p id="terms">'
            html += "<small><b>Filter: </b>The particular filter in the sequence.<br>"
            html += "<b>Event FOV: </b>The size of the FOV (in arc-minutes) for UVOT event data.<br>"
            html += "<b>Image FOV: </b>The size of the FOV (in arc-minutes) for UVOT image data.<br>"
            html += "<b>Max. Exp. Time: </b>The maximum amount of time the snapshot will spend on the particular filter in the sequence.<br>"
            html += "<b>Weighting: </b>Ratio of time spent on the particular filter in the sequence.<br>"
            html += "<b>Comments: </b>Additional notes that may be useful to know.<br></small>"
            html += "</p>"
            return html
        else:
            return "No data"

    def validate(self):
        """Validate API submission before submit

        Returns
        -------
        bool
            Was validation successful?
        """
        # Check username and shared_secret are set
        if self.uvotmode is None:
            return False
        if not self.username or not self.shared_secret:
            self.status.error("username and shared_secret parameters need to be supplied.")
            return False

        if type(self._uvot) != int:
            try:
                # See if it's a hex string
                self.uvotmode = int(self.uvotmode, 16)
            except ValueError:
                self.status.error(f"Invalid UVOT mode: {self.uvotmode}.")
                return False
        return True


# Aliases that are more PEP8 compliant
UVOTMode = Swift_UVOTMode
UVOTModeEntry = Swift_UVOTModeEntry
# Backwards compatibility names
UVOT_mode_entry = Swift_UVOTModeEntry
UVOT_mode = Swift_UVOTMode



"""
Module for astronomers using the Neil Gehrels Swift Observatory.

The `swifttools` module provides various features and functionality
designed to support users of the Swift satellite. The top-level
module has no contents, instead you will need to use one of the
sub-modules:

  * swifttools.swift_too -- Tools for submitting ToOs, and getting
    visibility and observing history.
  * xrt_prods -- Tools for the automated analysis of XRT Data.
"""

from .version import __version__



from .api_common import TOOAPI_Baseclass
from .api_daterange import TOOAPI_Daterange
from .api_status import TOOStatus
from .swift_clock import TOOAPI_ClockCorrect


class Swift_SAAEntry(TOOAPI_Baseclass, TOOAPI_ClockCorrect):
    """Simple class describing the start and end time of a Swift SAA passage.
     Attributes
    ----------
    begin : datetime
        Start time of the SAA passage
    end : datetime
        End time of the SAA passages
    """

    # API details
    api_name = "Swift_SAA_Entry"
    # Returned values
    _attributes = ["begin", "end"]
    # Display names of columns
    _varnames = {"begin": "Begin", "end": "End"}

    def __init__(self):
        # Attributes
        self.begin = None
        self.end = None
        # Internal values
        self._isutc = True

    @property
    def _table(self):
        header = [self._header_title("begin"), self._header_title("end")]
        data = [[self.begin, self.end]]
        return header, data

    @property
    def table(self):
        return ["begin", "end"], [[self.begin, self.end]]


class Swift_SAA(TOOAPI_Baseclass, TOOAPI_Daterange, TOOAPI_ClockCorrect):
    """Class to obtain Swift SAA passage times. Two versions are available: The
    Spacecraft definition (default) or an estimate of when the BAT SAA flag is
    up. Note that the BAT SAA flag is dynamically set based on count rate, so
    this result only returns an estimate based on when that is likely to happen.

    Attributes
    ----------
    entries : list
        Array of Swift_SAAEntry classes containing the windows.
    status : TOOStatus
        Status of API request
    """

    # API details
    api_name = "Swift_SAA"
    # Arguments
    _parameters = ["username", "begin", "end", "bat"]
    # Local parameters
    _local = ["shared_secret", "length"]
    # Returned Values
    _attributes = ["entries", "status"]
    # Returned classes
    _subclasses = [Swift_SAAEntry, TOOStatus]

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        begin : datetime
            Start of the period for which to fetch SAA passages
        end : datetime (optional)
            End of the period for which to fetch SAA passages
        length : int (default: 1)
            Number of days to calculate for
        bat : boolean
            If set to `True`, use BAT calculation for SAA passages, otherwise,
            use spacecraft.
        username : str (default 'anonymous')
            TOO API username.
        shared_secret : str (default 'anonymous')
            TOO API shared secret
        """
        # Attributes
        self.begin = None
        self.end = None
        self.length = 1
        self.bat = False
        # parse arguments
        self._parseargs(*args, **kwargs)
        self.status = TOOStatus()
        # Returned values
        self.entries = None
        # Internal values
        self._isutc = True

        # Submit if enough parameters are passed to the constructor
        if self.validate():
            self.submit()
        else:
            self.status.clear()

    def __getitem__(self, index):
        return self.entries[index]

    def __len__(self):
        return len(self.entries)

    @property
    def _table(self):
        if self.entries is None:
            return [], []
        else:
            vals = list()
            for i in range(len(self.entries)):
                header, values = self.entries[i]._table
                vals.append([i] + values[0])
            return ["#"] + header, vals

    def validate(self):
        """Validate API submission before submit

        Returns
        -------
        bool
            Was validation successful?
        """
        if self.begin is not None and self.end is not None:
            return True
        return False


# Alias
SAA = Swift_SAA
Swift_SAA_Entry = Swift_SAAEntry
SAAEntry = Swift_SAAEntry



from .api_status import TOOStatus
from .swift_calendar import Calendar
from .swift_clock import Clock
from .swift_data import Data
from .swift_guano import GUANO
from .swift_obsquery import AFST
from .swift_planquery import PPST
from .swift_requests import TOORequests
from .swift_saa import SAA
from .swift_uvot import UVOTMode
from .swift_visquery import VisQuery


class QueryJob(TOOStatus):
    """Class that enables fetching the results of already submitted jobs.
    Essentially the same as TOOStatus other than if the process has
    been completed the result of the job will be attached as `result`.

    Attributes
    ----------
    jobnumber : int
        TOO API job number
    username : str
        username for TOO API (default 'anonymous')
    shared_secret : str
        shared secret for TOO API (default 'anonymous')
    status : str
        status of API request
    timestamp : datetime
        time request was submitted
    began : datetime
        time request began processing
    completed : datetime
        time request finished processing
    errors : list
        list of error strings assoicated with the request
    warnings : list
        list of warning strings associated with the request
    too_id : list
        For a TOO request, the TOO ID assigned to a new request
    fetchresult : boolean
        Fetch the result of the query
    result : <varies>
        The completed job object
    """

    # These are the kinds of results that can be returned from a QueryJob
    _subclasses = [
        TOOStatus,
        VisQuery,
        AFST,
        UVOTMode,
        PPST,
        TOORequests,
        Calendar,
        GUANO,
        Data,
        Clock,
        SAA,
    ]
    # API name
    # This is really just a Swift_TOOStatus request with a twist
    api_name = "Swift_TOO_Status"

    def __init__(self, *args, **kwargs):
        """
        Parameters
        ----------
        jobnumber : int
            TOO API job number
        username : str
            username for TOO API (default 'anonymous')
        shared_secret : str
            shared secret for TOO API (default 'anonymous')
        """
        # Call the TOOStatus constructor with fetchresult=True to that make this a QueryJob
        TOOStatus.__init__(self, *args, fetchresult=True, api_name=self.api_name, **kwargs)

    @property
    def _table(self):
        """Table of request details"""
        _parameters = self._parameters + self._attributes
        table = [
            [row, getattr(self, row)]
            for row in _parameters
            if getattr(self, row) is not None and getattr(self, row) != "" and row != "result"
        ]
        table.append(["result", self.result.__class__.__name__ + " object"])
        return ["Parameter", "Value"], table
